/* SHOWPART.C -- Partition Table display program
   Copyright 1999-2004, Burton Systems Software.  All Rights Reserved.
   by David A. Burton   <dburton@burtonsys.com>
*/
char copr[] = "Copyright 1999-2004, Burton Systems Software.   http://www.burtonsys.com/\n";

/*-=>keyflag<=-- "&(#)%n, version %v, %d " */
/*  "&(#)SHOWPART.C, version 55, 31-Aug-04 " */

/*-=>keyflag<=-- "%v [%d]" */
char version[] = "55 [31-Aug-04]";

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <bios.h>
#include <dos.h>

#ifndef __TURBOC__
  #error "This must be compiled with Turbo-C"
  /* Note: if you convert to another C dialect, then be sure to use a #pragma
     to force the partition table structures to be packed to byte boundaries */
#endif

typedef void far * farvoidptr_t;
typedef enum {false_c=0, true_c=1} boolean_t;
typedef unsigned char uchar_t;
typedef unsigned int ushort_t;
typedef unsigned long ulong_t;
typedef uchar_t far * FarPtrByte_t;

char * savefilename;  /* filename from -F or -Sfilename command-line option */
FILE * savefile;
char * loadfilename;  /* filename from -W or -Tfilename command-line option */
FILE * loadfile;
boolean_t writeit;  /* true for -W option, false for -T option */
boolean_t fixit;  /* true for -F option, false for -S option */
boolean_t verbose;  /* true iff -V option specified */
boolean_t debugmode;  /* true iff -D option specified */

/* Look at the file handle table (usually in the PSP) to determine whether
   or not stderr and stdout are really the same file.  Return 1 if they are
   the same, 0 if they are different (because stdout has been redirected).
*/
int stderr_equals_stdout( void )
  {
#define STDIN  0   /* Handle for standard input  */
#define STDOUT 1   /* Handle for standard output */
#define STDERR 2   /* Handle for standard error  */
   char far * fhandle_table;
   static int result = -1;
    if (result == -1)
      {
        /* first, get real-mode address of the file handle table from PSP */
        fhandle_table = *(farvoidptr_t far *)MK_FP(_psp,0x34);
        /* then compare the table entries */
        result = (fhandle_table[STDOUT] == fhandle_table[STDERR]);
      }
    return result;
  } /* stderr_equals_stdout */


/* Read a character from the keyboard and return it in the low byte of the
   function result.  Extended codes have zero in the low byte and the extended
   code in the high byte.
*/
int GetChar( boolean_t flush )
  {
   int result;

    if (flush)
      {
        while (kbhit())
            (void)getch();
      }
    result = getch();
    if ((0 == result) || (0xE0 == result))
        result = getch() << 8;
    return result;
  } /* GetChar */


/*  get the answer to a yes/no question (he MUST enter Y or N)  */
boolean_t ask( char * question )
  {
   int ch;
   boolean_t result;
  again:
    fprintf( stderr, question );
    fflush(stderr);  /* force the prompt to display */
    ch = GetChar( /*flush=*/true_c );
    if ((ch=='Y') || (ch == 'y'))
        result = true_c;
    else if ((ch=='N') || (ch == 'n'))
        result = false_c;
    else
      {
        fprintf( stderr, "\nY or N, please:\n" );
        goto again;
      }
    fprintf( stderr, "%c\n", toupper(ch) );
    if (!stderr_equals_stdout())
      {
        printf( "%s%c\n", question, ch );
      }
    return result;
  } /* ask */


/*
My 1.08 Gig Quantum SCSI drive is formatted to Future Domain geometry:
63 sectors, 128 heads, 260 cylinders.  That should all be accessible
via the DOS Int 13H interface.

For drives with more than 1024 logical cylinders, an EBIOS is required
(or extender software, such as Disk Manager or EZ-Drive, which emulates
an EBIOS), except for an old-style "Type 3" BIOS which supports up to
4096 cylinders but only 63 heads.

This is from Ralf Brown's famous interrupt list:

|       CH = low eight bits of cylinder number
|       CL = sector number 1-63 (bits 0-5)
|            high two bits of cylinder (bits 6-7, hard disk only)
|       DH = head number
|       DL = drive number (bit 7 set for hard disk)
 . . .
|       the IBM AT BIOS and many other BIOSes use only the low four bits of
|         DH (head number) since the WD-1003 controller which is the standard
|         AT controller (and the controller that IDE emulates) only supports
|         16 heads
|       AWARD AT BIOS and AMI 386sx BIOS have been extended to handle more
|         than 1024 cylinders by placing bits 10 and 11 of the cylinder number
|         into bits 6 and 7 of DH

The old 4-bit head number limit is the source of the old 528 MB limit:
16 heads * 63 sectors * 1024 cylinders
                = 1,032,192 sectors
                = 516,096 KB (where 1 KB = 1024 bytes)
                = 528.482 MB (where 1 MB = 10**6 bytes)
                = 504 MB (where 1 MB = 2**20).

The Award/AMI solution mentioned above extends the maximum number
of cylinders to 4096, but limits the number of heads to 64 (or
maybe still 16?).  Hale Landis calls this a "BIOS Type 3" and says,
"Apparently this BIOS type is no longer produced by any BIOS vendor"
in chstran2.txt (version 4B, 6/24/99).

(Note: atafq-12.html says that this extension for 12-bit cylinder
numbers using the upper 2 bits of DH "was never widely implemented
and isn't supported anywhere."  According to that document, it is
much more common to simply use the entire 8-bit DH register for a
head number, allowing up to 256 logical heads (although limiting it
to 255 heads is the norm).  That document is part of the EIDE/ATA FAQ,
which can be found at http://thef-nym.sci.kun.nl/~pieterh/storage.html
or http://come.to/eide/ or various other locations.)

The Future Domain BIOS for my SCSI adapter, however, takes a
different (and incompatible!) approach.  It leaves the cylinder
limit at 1024, but allows 128 logical heads (or maybe 256?).
That should allow up to 4 GB (or maybe 8 GB?) maximum per drive,
but the partition granularity is nearly 4 megabytes per cylinder.

(I had a vague, and probably incorrect, memory that the Future
Domain controller maxxed out at 2 GB per drive rather than 4 GB ...
but that was apparently incorrect, since it worked with my 4.3 GB
Quantum SCSI hard drives.)

I once had an EIDE drive that was supposed to be 2,109,376 sectors,
which is 1.08 GB, but it formatted as 261 cylinders, which is
2,104,704 sectors, which means that I was wasting 4672 sectors
(2.4 MB) at the end, and wasting almost another 3 megabytes for my
oversized Boot Manager cylinder.)

One anomaly that I've noticed: the OS/2 Warp VDMs mask off the high
two bits of DH for Int 13H, AH=08 (get drive parameters).  Maybe
OS/2 noticed that my motherboard has an AMI BIOS?  But the BIOS
in question is the one on the SCSI card, not the motherboard!
So, the Future Domain BIOS reports DH=7F under DOS, but if the DOS
app is run in a VDM under OS/2 it sees DH=3F for the number of heads.
(This badly confuses all the other partition display & edit utilities
that I've tried to run in a DOS box under OS/2!)

Note that I think that there is the potential for a machine with
both EIDE and SCSI drives to use *both* of the INT 13 conventions
for large drives: 63*64*(up to 4096), and 63*128*(up to 1024).
                  \------EIDE------/      \--Future Domain--/

It would be almost miraculous if both kinds worked together in the
same machine with even 50% of the software that uses Int 13H, since
the high two bits of DH are being used for cylinder number on some
drives and head number on others.  Ugh!

                       SOME INT 13H FUNCTIONS
 ------------------------------------------------------------------
 cmd      command description                       notes
 --- ---------------------------------     -------------------------
  1  GET STATUS OF LAST DISK OPERATION     specify cmd & drive, rest 0s
  2  READ SECTOR(S) INTO MEMORY            specify all parameters
  3  WRITE DISK SECTOR(S)                     "     "      "
  4  VERIFY DISK SECTOR(S)                    "     "      "
  5  FLOPPY - FORMAT TRACK                 head should be 0 or 1
  5  FIXED DISK - FORMAT TRACK             nsects=interleave for XTs
  5  SEND SCSI MODE SELECT COMMAND         Future Domain SCSI BIOS
  6  FORMAT TRACK & SET BAD SECTOR FLAGS   XT/Port FIXED DISK
  6  FORMAT DRIVE WITH BAD SECTOR MAPPING  Future Domain SCSI BIOS
  6  IDENTIFY SCSI DEVICES                 Adaptec AHA-154xA/Bustek BT-542 BIOS
  7  FORMAT DRIVE STARTING AT GIVEN TRACK  XT/Port FIXED DISK
  7  FORMAT DRIVE                          Future Domain SCSI BIOS
  8  GET DRIVE PARAMETERS                  specify cmd & drive, rest 0s
 15h GET DISK TYPE
 1Bh GET POINTER TO SCSI DISK INFO BLOCK   Future Domain SCSI CONTROLLER
 41h EBIOS check extensions present        Fails without Extended BIOS
 42h EBIOS read sector(s)                  Requires Extended BIOS (LBA)
 43h EBIOS write sector(s)                    "        "      "     "
 48h EBIOS get drive parameters               "        "      "     "
*/

/* side-effect results of the my_biosdisk( 8, drive, 0,0,0,0,NULL ) */
int int13h8_maxsector = 63;  /* 8-63 */
int int13h8_maxhead = 0;  /* formerly 0-63, but can be 0..254 w/ modern BIOSs */
int int13h8_maxtrack = 0;  /* 0-1023, but can be larger w/ EIDE drives */
int unadjusted_int13h8_maxtrack = 0;  /* sometimes we increase int13h8_maxtrack by 1 to match the DPT; this is the original version */
int int13h8_highDHbits = 0;  /* 0-3 */
/*   If the drive has more than 64 heads, or more than 1024 cylinders, then
   int13h8_highDHbits will be non-zero.  It contains either bits 10-11 of
   the maximum track number (for an old "BIOS Type 3"), or else bits 6-7 of
   the maximum head number (for a modern BIOS).
     We examine the DPT to tell which it is, and adjust the int13h8_* values
   accordingly.
*/

/* true iff we have verified that LBA bios extensions are available for the
   current drive (EBIOS): */
boolean_t extended_BIOS;

boolean_t use_LBA_only;  /* currently ignored */

/* EBIOS Int 13h AH=48 (get disk parameters) results: */
ulong_t int13h48_maxtrack;  /* number of cylinders on disk, need to decrement by 1 */
ushort_t int13h48_maxhead;  /* number of heads, need to decrement by 1 */
ushort_t int13h48_maxsector;  /* number of sectors per track */
ulong_t int13h48_disksize;  /* number of sectors on entire disk */

FarPtrByte_t far * Int4xVect;  /* address of Int41H or Int46H vector */
FarPtrByte_t DPT_ptr;  /* if there's a DPT, this is just *Int4xVect */
ushort_t DPT_numcyl, DPT_numheads, DPT_numsec;  /* info from the DPT */
/* Side-effect of examine_DPT(), true iff Int13/Fn8 and the DPT were consistent
   except for the number of cylinders, and the DPT showed one more cylinder
   that Int13/Fn8 reported.  */
boolean_t Grew_NumCyl_By1 = false_c;
boolean_t Didnt_Grow_NumCyl_By1 = false_c;


/* number of sectors per logical cylinder */
ulong_t sect_per_cyl( void )
  {
    return int13h8_maxsector * (int13h8_maxhead+1LU);
  }


/* A "BIOS Type 3" supports up to 64 heads and up to 4096 cylinders.
   This type of BIOS is apparently no longer produced, sez Hale Landis.  */
boolean_t is_Type3( void )
  {
    if (   (int13h8_maxhead <= 63)
        && (int13h8_maxtrack > 1023)
        && (int13h8_maxtrack < 4096) )
      {
        return true_c;
      }
    else
        return false_c;
  } /* is_Type3 */


/* Compute track/head/sector number from absolute logical_sect (and
   int13h8_maxhead and int13h8_maxsector). */
void calc_T_H_S( ulong_t logical_sect, ushort_t * track, ushort_t * head, ushort_t * sect )
  {
   ulong_t temp;
    *track = logical_sect / sect_per_cyl();
    temp = logical_sect - ((*track) * sect_per_cyl());
    *head = temp / int13h8_maxsector;
    temp -= ((*head) * int13h8_maxsector);
    *sect = (ushort_t)(temp + 1);
  } /* calc_T_H_S */


ulong_t calc_LBA( ushort_t cylinder, ushort_t head, ushort_t sect )
  {
   ulong_t result;
    result = ((ulong_t)(sect - 1U)) + (head * (ulong_t)int13h8_maxsector)
               + (cylinder * sect_per_cyl());
    return result;
  } /* calc_LBA */


/* does DPT_ptr point to a "Translated DPT" with proper checksum? */
boolean_t is_translated_DPT( FarPtrByte_t DPT_ptr )
  {
   uchar_t cksum;
   int i;
    if (!DPT_ptr)
        return false_c;  /* there is no DPT */
    if ((DPT_ptr[3] & 0xF0) != 0xA0)  /* signature Axh? */
        return false_c;  /* not signature Ax */
    cksum = 0;
    for (i=0; i<=15; i++)
        cksum += DPT_ptr[i];
    if (0==cksum)
        return true_c;
    return false_c;  /* checksum wrong */
  } /* is_translated_DPT */


/* examine this drive's DPT, and perhaps fix up the int23h8_* globals */
void examine_DPT( int drive )
  {
   int DPT_err;
   ushort_t provisional_maxhead, provisional_maxtrack, bigger_maxhead, bigger_maxtrack;

    /* get info from the FDPT, for drives 80H and 81H, only */
    Int4xVect = NULL;
    DPT_ptr = NULL;
    DPT_numcyl = 0;
    DPT_numheads = 0;
    DPT_numsec = 0;
    DPT_err = 0;
    Didnt_Grow_NumCyl_By1 = Grew_NumCyl_By1 = false_c;
    if (0x80 == drive)
        /* first drive has DPT at int vector 41H */
        Int4xVect = MK_FP(0,4*0x41);
    else if (0x81 == drive)
        /* second drive has DPT at int vector 46H */
        Int4xVect = MK_FP(0,4*0x46);
    if (Int4xVect)
        DPT_ptr = *Int4xVect;
    if (DPT_ptr)
      {
        DPT_numcyl = *(ushort_t far *)DPT_ptr;
        DPT_numheads = DPT_ptr[2];
        DPT_numsec = DPT_ptr[14];
        if (0 == DPT_numheads)
            DPT_numheads = 256;
        if (DPT_numsec != int13h8_maxsector)
          {
            printf( "ERROR: int13 AH=8 returned max sect=%d, but DPT shows #sect=%d\n",
                    int13h8_maxsector, DPT_numsec );
            DPT_err++;
          }
      }

    /* if high 2 bits of DH are not zero, check for "BIOS Type 3" */
    provisional_maxhead = int13h8_maxhead;
    provisional_maxtrack = int13h8_maxtrack;
    bigger_maxhead = int13h8_maxhead | (int13h8_highDHbits << 6);  /* regs.h.dh */
    bigger_maxtrack = int13h8_maxtrack | (int13h8_highDHbits << 10);

    if (DPT_ptr)
      {
        if (DPT_numsec == int13h8_maxsector)
          {
            if ((DPT_numcyl > 1024) && (DPT_numheads <= 64))
              { /* looks like an old "BIOS Type 3" with track # 0..4095 */
                provisional_maxtrack = bigger_maxtrack;
              }
            else if (DPT_numheads > 64)
              { /* looks like a normal/modern BIOS, with up to 256 heads */
                provisional_maxhead = bigger_maxhead;
              }
          }
        if (DPT_numheads == (provisional_maxhead+1))
          {
            int13h8_maxhead = provisional_maxhead;
            if (0 == int13h8_highDHbits)
                printf( "High DH bits are not used.  MaxHead = %d.\n", int13h8_maxhead );
            else
                printf( "High DH bits are used for big head number, corrected MaxHead = %d.\n",
                        int13h8_maxhead );
            printf( "Good: Int13/Fn8 and DPT show consistent no. of heads (%d)\n",
                    DPT_numheads );
          }
        else
          {
            printf( "ERROR: Int13/Fn8 and DPT show inconsistent no. of heads!\n"
                    "Int13/Fn8 shows max head #%d, but DPT shows %d heads instead of %d\n",
                    provisional_maxhead, DPT_numheads, provisional_maxhead+1 );
            DPT_err++;
          }
        if (DPT_numcyl == (provisional_maxtrack+1))
          {
            unadjusted_int13h8_maxtrack = int13h8_maxtrack = provisional_maxtrack;
            if (0 != int13h8_highDHbits)
                printf( "BIOS Type 3 detected: High DH bits are used for big track number,\n"
                        "corrected maximum cylinder=%d.\n", int13h8_maxtrack );
            printf( "Good: Int13/Fn8 and DPT show consistent no. of cylinders (%d)\n",
                    DPT_numcyl );
          }
        else if (   (DPT_numcyl == (provisional_maxtrack+2))
                 && !DPT_err )
          {
            printf( "Note: The DPT shows one more cylinder than does Int13/Fn8.  When this happens\n"
                    "it might be because DiskManager or EZ Drive or a similar tool is using up\n"
                    "one cylinder.  Often, however, the last cylinder is just wasted, or perhaps\n"
                    "reserved for test purposes.  Most (but not all!) FDISKs make the conservative\n"
                    "choice and use the smaller (Int13/Fn8) number, often wasting the last 8.2 MB,\n" );
            if ((DPT_numcyl != 4097) && (DPT_numcyl != 1025))
              {
                printf( "but we assume the larger value.\n" );
                int13h8_maxtrack = provisional_maxtrack = DPT_numcyl - 1;
                /* note that unadjusted_int13h8_maxtrack is not changed here */
                /* not serious enough to consider it an error */
                Grew_NumCyl_By1 = true_c;
              }
            else
              {
                printf( "and so do we.\n" );
                Didnt_Grow_NumCyl_By1 = true_c;
              }
          }
        else
          {
            printf( "Warning: Int13/Fn8 and DPT show inconsistent no. of cylinders!\n"
                    "Int13/Fn8 shows max track #%d, but DPT shows %d cylinders instead of %d.\n",
                    provisional_maxtrack, DPT_numcyl, provisional_maxtrack+1 );
            if (    is_translated_DPT(DPT_ptr) && (1023==provisional_maxtrack)
                 && (DPT_numcyl > 1024) && extended_BIOS && !DPT_err )
              { /* Modern large disk EBIOS, for more than 1024 cylinders */
                printf( "EBIOS available, using DPT geometry.\n" );
                int13h8_maxsector = DPT_numsec;
                int13h8_maxhead = provisional_maxhead = DPT_numheads - 1;
                int13h8_maxtrack = provisional_maxtrack = DPT_numcyl - 1;
              }
            else
                DPT_err++;
          }

        if (DPT_err)
          {
           char * question;
            printf( "Either Int13/Fn8 or the DPT is wrong!\n" );
            if (is_translated_DPT(DPT_ptr) || (_osmajor >= 20))
                /* _osmajor==20 means OS/2, which has buggy Int13/Fn8 in DOS boxes;
                   if is_translated_DPT() returns true then the DPT checksum
                   was right, so it is pretty reliable.  */
                question = "Believe the DPT instead of Int13/Fn8?  (the right answer is probably Y)";
            else
                /* DPTs are considered obsolete */
                question = "Believe the DPT instead of Int13/Fn8?  (the right answer is probably N)";
            if (ask(question))
              {
                int13h8_maxsector = DPT_numsec;
                int13h8_maxhead = provisional_maxhead = DPT_numheads - 1;
                int13h8_maxtrack = provisional_maxtrack = DPT_numcyl - 1;
                DPT_err = 0;
              }
          }
      }
    else
      {
        /* There's no DPT.  This may be the 3rd drive.  Probably not a Type 3 BIOS. */
        printf( "There is no DPT for this drive.\n" );
        provisional_maxhead = bigger_maxhead;
        DPT_err = -1;
      }
    if (!DPT_err)
      {
        int13h8_maxhead  = provisional_maxhead;
        int13h8_maxtrack = provisional_maxtrack;
      }
    if (int13h8_highDHbits && DPT_err)
      { /* if high 2 bits of DH are not zero, check for "BIOS Type 3" */
        printf( "I could not determine for sure whether this is an old \"Type 3\" disk BIOS.\n"
                "My best guess is that it is" );
        if ((provisional_maxhead > 63) || (provisional_maxtrack < 1023))
            printf( " not" );
        printf( ".\n" );
        printf( "Old \"Type 3\" BIOSs support drives with up to 4096 cylinders & up to 64 heads.\n"
                "Normal BIOSs support drives with up to 1024 cylinders & up to 255 or 256 heads.\n" );
        printf( "Both support drives of up to about 8.4 GB.  (EBIOS/LBA is needed for >8.4 GB.)\n" );
        printf( "This drive's geometry (C/H/S) is %d/%d/%d (normal) or %d/%d/%d (Type 3).\n",
                int13h8_maxtrack+1, bigger_maxhead+1, int13h8_maxsector,
                bigger_maxtrack+1, int13h8_maxhead+1, int13h8_maxsector );
        if (ask("Is this a \"BIOS Type 3\" (big cylinder number)?"))
            unadjusted_int13h8_maxtrack = int13h8_maxtrack = bigger_maxtrack;
        else
            int13h8_maxhead = bigger_maxhead;
        printf( "Corrected drive geometry is:\n"
                "Maxsector = %d, maxhead = %d, maxtrack = %d\n\n",
                int13h8_maxsector, int13h8_maxhead, int13h8_maxtrack );
      }
  } /* examine_DPT */


/* 16-byte Device Address Packet structure used for Int 13h, AH=41h-48h */
typedef struct {
  uchar_t pkt_size;  /* always 16 */
  uchar_t zeropad1;  /* 0 */
  uchar_t num_blks_to_xfer;  /* 0..127 */
  uchar_t zeropad2;  /* 0 */
  farvoidptr_t bufadr;  /* address of transfer buffer, seg:ofs */
  ulong_t starting_LBA;  /* starting logical block address */
  ulong_t highzero;  /* 0 until drives exceed 2199 GB */
} DevAdrPacket_t;


/* Does BIOS support the "T13" LBA fixed disk access extensions?  Ask it.
   (See http://www.microsoft.com/ddk/ddkdocs/win98ddk/storage_60qe.htm) */
boolean_t check_extensions_present_41h( int drive )
  {
   union REGS regs;
   struct SREGS sregs;
    segread( &sregs );
    memset( &regs, 0, sizeof regs );
    regs.h.ah = 0x41;
    regs.h.dl = drive;
    regs.x.bx = 0x55AA;
    int86x( 0x13, &regs, &regs, &sregs );
    if (regs.x.cflag || (regs.x.bx != 0xAA55) || (regs.h.ah < 0x21))
        return false_c;
    if (regs.x.cx & 1)
        return true_c;
    return false_c;
  } /* check_extensions_present_41h */


int my_biosdisk( int cmd, /* e.g., 0x02 for "Read Sector(s)" */
                 int drive, /* 0x80 for first hard drive, etc.. */
                 int head, /* 0-254 (modern BIOSs), 0-63 (AMI), 0-15 (trad.) */
                 int track, /* cylinder: 0-1023 (trad.), 0-4095 (AMI) */
                 int sector, /* 1-63 */
                 int nsects,
                 void far *buffer )
  {
   union REGS regs;
   struct SREGS sregs;
   int result;
   char tmp[512];

    if ((sector & ~0x3F) || (drive & 0xFF70))
      {
        printf( "ERR: bad parameter to my_biosdisk(,0x%04X,,,0x%04X,,)\n",
                drive, sector );
        exit(2);
      }
    segread( &sregs );
    memset( &regs, 0, sizeof regs );
    regs.h.ah = cmd;
    regs.h.dl = drive;
    if (!buffer)
      {
        buffer = tmp;
        if (nsects)
            nsects = 1;
      }
    regs.x.bx = FP_OFF( buffer );
    sregs.es = FP_SEG( buffer );
    if (8==cmd)
      {
        int13h8_maxsector = 0;
        int13h8_maxhead = 0;
        int13h8_maxtrack = 0;
        unadjusted_int13h8_maxtrack = 0;
        int13h8_highDHbits = 0;
      }
    else
      {
        regs.h.al = nsects; /* no. of sectors, for functions 2, 3 and 4 */
        regs.h.ch = track & 0xFF;  /* bits 0-7 of cylinder number */
        regs.h.cl = (sector & 0x3F) | ((track & 0x0300) >> 2);
                    /* CL is sector number and bits 8-9 of cylinder number */
        regs.h.dh = head | ((track & 0x0C00) >> 4);
      }
    /* Note: the original Int 13H interface supported drives of at most 528 MB,
       because the maximum head number had to be 0-15.  Now, however, there
       are at least two approaches in use which extend this.  One approach
       simply allows the head number to be 0-127 or 0-255; that approach is
       used by Future Domain SCSI controllers and modern BIOSs.  The other
       approach (used by some older AMI & Award BIOSs, which Hale Landis calls
       "BIOS Type 3") allows a head number of 0-63, but also uses the high two
       bits of DH to specify bits 10-11 of a 12-bit track number; that lets the
       track number range from 0-4095 (with various bits scattered over three
       different 8-bit registers).  The above calculation should work for
       either approach.
    */
    int86x( 0x13, &regs, &regs, &sregs );
    result = regs.x.cflag ? regs.x.ax : 0;
    if ((!result) && (8==cmd))
      {
        /* get info from Int 13, AH=8 */
        int13h8_maxsector = (regs.h.cl & 0x3F);
        int13h8_maxhead = (regs.h.dh & 0x3F);
        int13h8_maxtrack = regs.h.ch + ((((int)regs.h.cl) << 2) & 0x300);
        unadjusted_int13h8_maxtrack = int13h8_maxtrack;
        int13h8_highDHbits = ((regs.h.dh >> 6) & 3);
      }
    return result;
  } /* my_biosdisk */


/* read or write sector(s) using EBIOS's Logical Block Address (LBA) functions */
int my_ebiosdisk( int cmd, /* 0x42 for "Read Sector(s)" or 0x43 for "Write Sector(s)" */
                  int drive, /* 0x80 for first hard drive, etc.. */
                  ulong_t lba_msw, /* 0L */
                  ulong_t lba_lsw, /* sector number 0-... (LBA) */
                  int nsects,
                  void far *buffer )
  {
   union REGS regs;
   struct SREGS sregs;
   int result;
   DevAdrPacket_t pkt;
   char tmp[512];

    if ((cmd < 0x40) || (drive & 0xFF70) || lba_msw)
      {
        printf( "ERR: bad parameter to my_ebiosdisk(0x%02X,0x%02X,%d,,,)\n",
                cmd, drive, lba_msw );
        exit(2);
      }
    if (!extended_BIOS)
      {
        printf( "ERROR: No EBIOS, LBA disk access unavailable\n" );
        exit(2);
      }
    if (!buffer)
      {
        buffer = tmp;
        if (nsects)
            nsects = 1;
        memset( tmp, 0, 512 );
      }
    segread( &sregs );
    memset( &regs, 0, sizeof regs );
    memset( &pkt, 0, sizeof(DevAdrPacket_t) );
    regs.h.ah = cmd;
    regs.h.dl = drive;
    regs.x.si = FP_OFF( &pkt );
    sregs.ds = FP_SEG( &pkt );
    pkt.pkt_size = sizeof(DevAdrPacket_t);  /* 16 */
    pkt.num_blks_to_xfer = nsects;
    pkt.bufadr = buffer;
    pkt.starting_LBA = lba_lsw;
    pkt.highzero = lba_msw;

    int86x( 0x13, &regs, &regs, &sregs );

    result = regs.x.cflag ? regs.h.ah : 0;
    return result;
  } /* my_ebiosdisk */


void dump_a_sector( char far * bufr );

int my_disk_rd( /* Uses either 0x02 or 0x42 for "Read Sector(s)" */
                 int drive, /* 0x80 for first hard drive, etc.. */
                 int head, /* 0-254 (modern BIOSs), 0-63 (AMI), 0-15 (trad.) */
                 unsigned int track, /* cylinder: 0-1023 (trad.), 0-4095 (AMI) */
                 int sector, /* 1-63 */
                 int nsects,
                 void far *buffer )
  {
   int result;
    if ((track > 1023) && extended_BIOS)
      {
        result = my_ebiosdisk( 0x42, drive, 0L, calc_LBA(track,head,sector),
                               nsects, buffer );
      }
    else
        result = my_biosdisk( 0x02, drive, head, track, sector,
                              nsects, buffer );
    if (debugmode)
      {
        printf("---------- begin dbg dmp ----------\n");
        printf("my_disk_rd(%02X,%d,%u,%d):\n", drive, head, track, sector );
        dump_a_sector(buffer);
        printf("----------- end dbg dmp -----------\n");
      }
    return result;
  } /* my_disk_rd */


int my_disk_wr( /* Uses either 0x03 or 0x43 for "Read Sector(s)" */
                 int drive, /* 0x80 for first hard drive, etc.. */
                 int head, /* 0-254 (modern BIOSs), 0-63 (AMI), 0-15 (trad.) */
                 unsigned int track, /* cylinder: 0-1023 (trad.), 0-4095 (AMI) */
                 int sector, /* 1-63 */
                 int nsects,
                 void far *buffer )
  {
   int result;
    if (debugmode)
      {
        printf("---------- begin dbg dmp ----------\n");
        printf("my_disk_wr(%02X,%d,%u,%d):\n", drive, head, track, sector );
        dump_a_sector(buffer);
        printf("----------- end dbg dmp -----------\n");
      }
    if ((track > 1023) && extended_BIOS)
      {
        result = my_ebiosdisk( 0x43, drive, 0L, calc_LBA(track,head,sector),
                               nsects, buffer );
      }
    else
        result = my_biosdisk( 0x03, drive, head, track, sector,
                              nsects, buffer );
    return result;
  } /* my_disk_wr */


struct H_S_C_s {
    uchar_t head, sector, cyl;
  };
struct partrec_s {
    uchar_t boot_ind;
    struct H_S_C_s s;  /* starting head/sector/cylinder */
    uchar_t sys_ind;
    struct H_S_C_s e;  /* ending head/sector/cylinder */
    ulong_t relative_sect;
    ulong_t number_of_sect;
  };
struct parttable_s {
    struct partrec_s part[4];
    ushort_t AA55H; /* signature word, should always be 0xAA55 */
  } * pt;
/* While examining a partition table, we first read the sector into a 512-byte
   buffer, then we set pt to point to offset 0x1BE in that buffer.  Then we
   can use the above structure definitions to access the various partition
   table fields via *pt.
*/

/* where is currently-being-displayed partition table located on the disk;
   side-effect of calling display_partition_table(); setting these globals
   is the first thing that display_partition_table() does.  */
ushort_t this_ptbl_at_cyl, this_ptbl_at_head, this_ptbl_at_sect;
/* start of displayed partition table, as absolute logical sector number,
   calculated from the drive/head/track/sector input parameters to
   display_partition_table(), and set as a side-effect: */
ulong_t partition_logical_sect = 0;
/* locations of other interesting extended partition table records */
ulong_t prev_ptbl_at_LBA = 0;
ulong_t start_of_first_extended = 0;

/* these are more side-effects of display_partition_table */
int partition_table_bad = 0;  /* "1" if we decide that partition table is bad */
ushort_t biggest_head = 0;
ushort_t biggest_sector = 0;
ushort_t biggest_track = 0;
ushort_t extended_head = 0;
ushort_t extended_sector = 0;
ushort_t extended_track = 0;
ulong_t extended_relative_sect = 0;
int num_boot_indicators_seen = 0;
int num_partitions = 0;


/*   Under OS/2, "Get Drive Parameters" returns the wrong value in DH, which is
   supposed to be the maximum head number.  The high two bits are zeroed out.
   We work around that by examining the DPT (we used to examine the partition
   table), but I fretted that the head number might be mangled in the other
   direction, too: when reading and writing the drive via Int 13H functions
   2 and 3.
     Fortunately, that does not seem to be the case, but here is where we make
   sure.  (I simply verify that head#0/sector#1 and head#64/sector#1 of track
   zero contain different data.)
*/
void check_for_headnum_truncation( int drive )
  {
   int rc;
   int cntr;
   boolean_t equal;
   int sectnum;
   char buf1[512], buf2[512];
    printf( "Checking for head-number truncation:\n" );
    rc = my_biosdisk( /*cmd=*/2, drive, /*head=*/0, /*track=*/0, /*sect=*/1,
                      /*nsects=*/1, buf1 );
    if (rc)
      {
        printf( "ERROR: Read failed, my_biosdisk(2,0x%X,0,0,1,1,)=%d\n",
                drive, rc );
        exit(1);
      }
    rc = my_biosdisk( /*cmd=*/2, drive, /*head=*/64, /*track=*/0, /*sect=*/1,
                      /*nsects=*/1, buf2 );
    if (rc)
      {
        printf( "ERROR: Read failed, my_biosdisk(2,0x%X,64,0,1,1,)=%d\n",
                drive, rc );
        exit(1);
      }
    equal = true_c;
    for (cntr=0; cntr<512; cntr++)
      {
        if (buf1[cntr] != buf2[cntr])
            equal = false_c;
      }
    if (equal)
      {
        printf( "Warning: reads don't work for head numbers > 63.\n" );
        printf( "Try running under DOS instead of OS/2.\n" );
        if (!ask("Keep going anyhow?  (There's probably no point!) (Y/N)"))
            exit(1);
      }
    else
        printf( "Good news: head numbers don't seem to be truncated for reads.\n" );
  } /* check_for_headnum_truncation */


/* given a 16-byte partition table record, calculate the starting and ending
   cylinder, head and sector numbers (just a bit of shift & fiddle code) */
void split_CHSs( struct partrec_s * prec,
                 ushort_t * cylX, ushort_t * headX, ushort_t * sectX,
                 ushort_t * cylY, ushort_t * headY, ushort_t * sectY )
  {
    *sectX = prec->s.sector & 0x3F;
    *cylX = prec->s.cyl + ((prec->s.sector << 2) & 0x300);
    *headX = prec->s.head;
    *sectY = prec->e.sector & 0x3F;
    *cylY = prec->e.cyl + ((prec->e.sector << 2) & 0x300);
    *headY = prec->e.head;
    if (is_Type3())
      { /* Oops, top two bits of head numbers are really bits 10 & 11
           of the cylinder number.  Fix 'em up.  */
        *cylX |= (((ushort_t)(prec->s.head & 0xC0)) << 4);
        *headX &= 0x3F;
        *cylY |= (((ushort_t)(prec->e.head & 0xC0)) << 4);
        *headY &= 0x3F;
      }
  } /* split_CHSs */


/* Examine the partition type / system indicator byte; if it is an LBA-only
   partition type, then return true.  For type 0Dh, we make the determination
   based on whether the starting CHS value is FF/FF/FF or FF/FE/FF (though
   I'm not sure that's correct).  */
boolean_t is_LBA_only( struct partrec_s * prec )
  {
    if (   (prec->sys_ind == 0x0C)
        || (prec->sys_ind == 0x0E)
        || (prec->sys_ind == 0x0F)
        || (prec->sys_ind == 0x1F)
        || (prec->sys_ind == 0x82)
        || (prec->sys_ind == 0x83)
        || (prec->sys_ind == 0x85)
        || (prec->sys_ind == 0x93) )
       return true_c;
    if (prec->sys_ind == 0x0D)
      {
        if (  (0xFF == prec->s.sector)
            && (0xFF == prec->s.cyl)
            && ((0xFF == prec->s.head) || (0xFE == prec->s.head)) )
           return true_c;
      }
    return false_c;
  } /* is_LBA_only */


/* Return a string name for the partition type.  Many of these are from one
   of Hale Landis' invaluable "How It Works" documents, parttab2.txt.  See also
   http://www.win.tue.nl/math/dw/pp/aeb/partitions/partition_types-1.html or
   run "sfdisk -T" under Linux. */
char * partitiontype( uchar_t sys_ind )
  {
   char * result;
    result = "?";
    switch (sys_ind)
      {
       case 0: result = "none"; break;
       case 1: result = "FAT12"; break;
       case 2:
       case 3: result = "Xenix"; break;
       case 4: result = "FAT16 <32MB"; break;
       case 5: result = "Extd"; break;
       case 6: result = "FAT16"; break;
       case 7: result = "HPFS/NTFS/QNX/etc."; break;
       case 8: result = "OS/2 1.x HPFS, DELL, AIX"; break;
       case 9: result = "AIX or Coherent"; break;
       case 10 /*0A*/: result = "OS/2 Boot Mgr"; break;
       case 11 /*0B*/: result = "FAT32"; break;
       case 12 /*0C*/: result = "FAT32-LBA"; break;
       case 13 /*0D*/: result = "Type7-LBA?"; break;
       case 14 /*0E*/: result = "FAT16-LBA"; break;
       case 15 /*0F*/: result = "Extd-LBA"; break;
       case 0x11: result = "Hidden FAT12"; break;
       case 0x14: result = "Hidden FAT16 <32MB"; break;
       case 0x15: result = "Hidden Extd"; break;
       case 0x16: result = "Hidden FAT16"; break;
       case 0x17: result = "Hidden HPFS/NTFS/etc."; break;
       case 0x1F: result = "Hidden Extd-LBA"; break;
       case 0x50: result = "OnTrack DM (RO)"; break;
       case 0x51: result = "Novell, or OnTrack (RO/WO)"; break;
       case 0x52: result = "CP/M or Microport"; break;
       case 0x53: result = "OnTrack DM (WO)"; break;
       case 0x54: result = "OnTrack DM (DDO)"; break;
       case 0x55: result = "EZ Drive"; break;
       case 0x56: result = "GoldenBow VFeature"; break;
       case 0x61: result = "SpeedStore"; break;
       case 0x63: result = "Mach, GNU HURD, or Unix Sys V/386"; break;
       case 0x64: result = "Novell or SpeedStor"; break;
       case 0x65: result = "Novell 286 Netware"; break;
       case 0x66: result = "Novell 386 Netware"; break;
       case 0x67: result = "Novell"; break;
       case 0x68: result = "Novell"; break;
       case 0x69: result = "Novell"; break;
       case 0x75: result = "PC/IX"; break;
       case 0x80: result = "Minix"; break;
       case 0x81: result = "Linux or Minix"; break;
       case 0x82: result = "Linux swap, Prime, or Solaris"; break;
       case 0x83: result = "Linux ext2fs"; break;
       case 0x85: result = "Extd-Linux"; break;
       case 0x86: result = "NTFS volume set"; break;
       case 0x87: result = "HPFS FT mirrored"; break;
       case 0x92: result = "Linux swap"; break;
       case 0x93: result = "Linux or Ameba"; break;
       case 0x94: result = "Ameba bad blk tbl"; break;
       case 0x99: result = "Mylex EISA SCSI"; break;
       case 0xA0: result = "SaveToDisk"; break;
       case 0xA5: result = "FreeBSD"; break;
       case 0xA9: result = "NetBSD"; break;
       case 0xB7:
       case 0xB8: result = "BSDI"; break;
       case 0xBE: result = "Solaris boot"; break;
       case 0xC1: result = "DR-DOS FAT12"; break;
       case 0xC4: result = "DR-DOS FAT16"; break;
       case 0xC6: result = "DR-DOS Huge"; break;
       case 0xC7: result = "Cyrnix boot or HPFS FT disabled mirrored"; break;
       case 0xD8: result = "CP/M 86"; break;
       case 0xDB: result = "Concurrent DOS, CP/M, CTOS"; break;
       case 0xE1: result = "SpeedStore"; break;
       case 0xE3: result = "Storage Dimensions"; break;
       case 0xE4: result = "SpeedStore"; break;
       case 0xEB: result = "BeOS BFS"; break;
       case 0xF1: result = "Storage Dimensions"; break;
       case 0xF2: result = "DOS 3.3+ second partition"; break;
       case 0xF4: result = "SpeedStore or Storage Dimensions"; break;
       case 0xFE: result = "PS/2 IMP, LANstep"; break;
       case 0xFF: result = "Xenix bad block tbls"; break;
       default: result = "????"; break;
      }
    return result;
  } /* partitiontype */


/* true iff sys_ind is a "system indicator" (partition type byte) which
   indicates an extended partition table location */
boolean_t is_extd_type( uchar_t sys_ind )
  {
    /* 05 means "regular" extended partition (OS/2 Warp Connect 3.0 &
          Partition Magic create this, even above 8.4 GB)
       0F means LBA-only extended partition table (Win95 OSR2 creates
          this in the MBR, but apparently not anywhere else)
       85 means Linux-specific LBA-only extended partition.
       15 means hidden 05.
       1F means hidden 0F.
       I think that these are all relative to the location of the FIRST
       extended partition table (definitely true for 05 and 0F, probably
       also true for 85). */
    if (   (sys_ind == 5) || (sys_ind == 0x0F) || (sys_ind == 0x85)
        || (sys_ind == 0x15) || (sys_ind == 0x1F) )
         return true_c;
     else
         return false_c;
  } /* is_extd_type */


/* return true for an empty partition table slot */
boolean_t all_zeros( struct partrec_s *prec )
  {
    return !(   prec->boot_ind || prec->sys_ind
             || prec->relative_sect || prec->number_of_sect
             || prec->s.head || prec->s.sector || prec->s.cyl
             || prec->e.head || prec->e.sector || prec->e.cyl );
  } /* all_zeros */


/*   Given a partition record, and the absolute LBA of the partition table
   sector, and the absolute starting LBA of the first extended partition
   table sector, return the absolute starting LBA of the partition which is
   described by this partition record.
     For partition records in the first partition table sector,
   partition_logical_sect should be 0.
     For partition records in the 2nd partition table sector,
   partition_logical_sect and start_of_first_extended should be equal but
   non-zero.  */
ulong_t calc_logical_start_sect( struct partrec_s * prec,
                                 int table_number, /* 1 for MBR, 0 for first extended, etc. */
                                 ulong_t partition_logical_sect, /* LBA of partition table sector */
                                 ulong_t start_of_first_extended /* LBA of first ext'd partition table */
                               )
  {
   ulong_t abs_start_LBA;  /* result */
    abs_start_LBA = prec->relative_sect;
    if (   ((1 == table_number) && partition_logical_sect)
        || ((2 == table_number) && (partition_logical_sect != start_of_first_extended))
        || ((table_number > 1) && ((!partition_logical_sect) || !start_of_first_extended))
        )
      {
        printf( "ERR: table_number=%d, partition_logical_sect=%lu, start_of_first_extended=%lu\n",
                table_number, partition_logical_sect, start_of_first_extended );
      }
    if (abs_start_LBA && (table_number > 1))
      { /* In the master boot record (in sector 1, track 0, head 0 -> LBA 0),
           the logical sector numbers are absolute sector numbers.  In the
           first extended partition table sector, the logical sector numbers
           are relative to that partition table sector.  In subsequent
           extended partition table sectors, what the logical sector number
           is relative-to depends upon the type of the partition table slot
           in question.  For type 05/0F/85, it is the location of another
           extended partition table sector, and its logical sector number is
           relative to the sector number of the FIRST extended partition table
           sector; other types have logical sector numbers that are relative
           to the CURRENT extended partition table sector.  What a mess! */
        if (is_extd_type(prec->sys_ind) && (table_number > 2))
          {
            abs_start_LBA += start_of_first_extended;
          }
        else
            /* all other partition table LBA locations are relative to location
               of CURRENT extended partition table */
          {
            abs_start_LBA += partition_logical_sect;
          }
      }
    return abs_start_LBA;
  } /* calc_logical_start_sect */


/* print_table_entry also sets table_entry_start_LBA as side-effect result */
ulong_t table_entry_start_LBA;

/* description of CHS large drive placeholder style */
char start_placeholder_style[50], end_placeholder_style[50];

/* Display one of the four partition table records in a boot sector.
   Return true iff this the start of the partition is LBA-only (CHS is
   a large drive placeholder).  Expects global partition_logical_sect
   to already be the start absolute LBA of displayed partition table */
boolean_t print_table_entry( FILE * outfile, int table_number, struct partrec_s * prec )
  {
   /* starting cylinder/head/sector numbers: */
   ushort_t cylX, headX, sectX, cylX2, headX2, sectX2;
   ulong_t logical_start_sect;  /* starting absolute LBA */
   /* ending cylinder/head/sector numbers: */
   ushort_t cylY, headY, sectY, cylY2, headY2, sectY2;
   ulong_t logical_end_sect;  /* ending absolute LBA */
   enum {identical_c=0, LBAonly_c=1, inconsistent_c=2} start_is_LBA_only, end_is_LBA_only;
   int i;

    /* first, dump the 16 bytes in hexidecimal */
    for (i=0; i<=15; i++)
      {
        fprintf( outfile, "%02X", ((uchar_t *)prec)[i] );
        /* add blanks to make it line up with caption2L[] */
        if ((i==0) || (i==3) || (i==4) || (i==7) || (i==11))
            fprintf( outfile, "  " );
        else if (i==15)
            fprintf( outfile, "\n" );
        else
            fprintf( outfile, " " );
      }
    if (all_zeros(prec))
      {
        table_entry_start_LBA = 0L;  /* side-effect result! */
        fprintf( outfile, "  (Unused partition table slot)\n" );
        return false_c;
      }
    num_partitions++;  /* side-effect result */

    fprintf( outfile, " BootInd=%02X  SysInd=%02X='%s'\n",
             prec->boot_ind, prec->sys_ind, partitiontype(prec->sys_ind) );
    if (0x80 == prec->boot_ind)
        num_boot_indicators_seen++;
    else if (prec->boot_ind != 0x00)
      {
        fprintf( outfile, "ERROR: BootInd should be 00 or 80H, but is actually %02XH\n",
                 prec->boot_ind );
      }

    split_CHSs( prec, &cylX, &headX, &sectX, &cylY, &headY, &sectY );

    logical_start_sect = calc_logical_start_sect( prec, table_number,
                             partition_logical_sect, start_of_first_extended );
    table_entry_start_LBA = logical_start_sect;  /* side-effect result! */

    logical_end_sect = prec->number_of_sect + logical_start_sect - 1LU;

    calc_T_H_S( logical_start_sect,
                &cylX2, &headX2, &sectX2 );  /* calculated starting CHS */
    calc_T_H_S( logical_end_sect,
                &cylY2, &headY2, &sectY2 );  /* calculated ending CHS */

    start_placeholder_style[0] = end_placeholder_style[0] = '\0';
    start_is_LBA_only = end_is_LBA_only = inconsistent_c;
    if (is_LBA_only( prec ))
        start_is_LBA_only = end_is_LBA_only = LBAonly_c;  /* this is an LBA-only partition */

    /* determine consistency for start LBA v. CHS */
    if (calc_LBA(cylX,headX,sectX) == logical_start_sect)
        start_is_LBA_only = identical_c;
    else
      {
        if ((headX==headX2) && (sectX==sectX2) && (cylX == (cylX2 & 1023)))
          { /* Win95 OSR2 style: just drop high bits of cylinder */
            start_is_LBA_only = LBAonly_c;
            strcpy( start_placeholder_style, "/MS" );
          }
        if ((headX==headX2) && (sectX==sectX2) && (cylX==1023))
          { /* Partition Magic style: CHS=1023/0/1 */
            start_is_LBA_only = LBAonly_c;
            strcat( start_placeholder_style, "/PowerQuest" );
          }
        if ((headX==headX2) && (sectX==sectX2) && (cylX==1022))
          { /* V-Communidations style: CHS=1022/0/1 */
            start_is_LBA_only = LBAonly_c;
            strcat( start_placeholder_style, "/V-Comm" );
          }
        if (((headX==255)||(headX==254)) && (sectX==63) && (cylX==1023))
          { /* FF FF FF or FE FF FF */
            start_is_LBA_only = LBAonly_c;
            if (headX==254)
                strcat( start_placeholder_style, "/IBM" );
            else
                strcat( start_placeholder_style, "/FFFFFF" );
          }
        /* else leave it as inconsistent_c (or LBAonly_c if is_LBA_only()==true_c) */
      }

    /* determine consistency for end LBA v. CHS */
    if (calc_LBA(cylY,headY,sectY) == logical_end_sect)
        end_is_LBA_only = identical_c;
    else
      {
        if ((headY==headY2) && (sectY==sectY2) && (cylY == (cylY2 & 1023)))
          { /* Win95 OSR2 style: just drop high bits of cylinder */
            end_is_LBA_only = LBAonly_c;
            strcpy( end_placeholder_style, "/MS" );
          }
        else if ((headY==headY2) && (sectY==sectY2) && (cylY==1023))
          { /* Partition Magic style: CHS=1023/254/63 */
            end_is_LBA_only = LBAonly_c;
            strcat( end_placeholder_style, "/PowerQuest" );
          }
        else if ((headY==headY2) && (sectY==sectY2) && (cylY==1022))
          { /* V-Communications style: CHS=1022/254/63 */
            end_is_LBA_only = LBAonly_c;
            strcat( end_placeholder_style, "/V-Comm" );
          }
        else if (((headY==255)||(headY==254)) && (sectY==63) && (cylY==1023))
          { /* FF FF FF or FE FF FF*/
            end_is_LBA_only = LBAonly_c;
            if (headX==254)
                strcat( end_placeholder_style, "/IBM" );
            else
                strcat( end_placeholder_style, "/FFFFFF" );
          }
        /* else leave it as inconsistent_c (or LBAonly_c if is_LBA_only()==true_c) */
      }

    /* display start-of-partition info */
    fprintf( outfile, "  begin: C=%u H=%u S=%u  LBA=%ld=%08lXh",
             cylX, headX, sectX, prec->relative_sect, prec->relative_sect );
    switch (start_is_LBA_only)
      {
        case identical_c:
          fprintf( outfile, "  (CHS & LBA are consistent)\n" );
          break;
        case LBAonly_c:
          if (start_placeholder_style[0])
            {
              start_placeholder_style[0] = ' ';  /* change first '/' to ' ' */
              strcat( start_placeholder_style, "-style" );
            }
          fprintf( outfile, " -> true C/H/S=%u/%u/%u\n"
                            "         Note: Stored start CHS is%s large drive placeholder.\n",
                   cylX2, headX2, sectX2, start_placeholder_style );
          break;
        case inconsistent_c:
          fprintf( outfile, "  LBA->C/H/S=%u/%u/%u  CHS->absLBA=%ld\n"
                            "         Error: Start CHS & LBA are inconsistent!\n",
                   cylX2, headX2, sectX2, calc_LBA(cylX,headX,sectX) );
          break;
      }

    /* display end-of-partition info */
    fprintf( outfile, "    end: C=%u H=%u S=%u  size=%ld (%ld MB)",
             cylY, headY, sectY, prec->number_of_sect,
             ((1024L+prec->number_of_sect) / 2048L) );
    switch (end_is_LBA_only)
      {
        case identical_c:
          fprintf( outfile, "  (size & CHS are consistent)\n" );
          break;
        case LBAonly_c:
          if (end_placeholder_style[0])
            {
              end_placeholder_style[0] = ' ';  /* change first '/' to ' ' */
              strcat( end_placeholder_style, "-style" );
            }
          fprintf( outfile, " -> true C/H/S=%u/%u/%u\n"
                            "         Note: Stored end CHS is%s large drive placeholder.\n",
                   cylY2, headY2, sectY2, end_placeholder_style );
          break;
        case inconsistent_c:
          fprintf( outfile, " -> calculated C/H/S=%u/%u/%u\n"
                            "         Warning: Calculated CHS inconsistent with ending CHS!\n",
                   cylY2, headY2, sectY2 );
          break;
      }
    return (start_is_LBA_only == LBAonly_c);
  } /* print_table_entry */


typedef struct saved_sector_s saved_sector_t;
typedef saved_sector_t * saved_sector_ptr_t;

struct saved_sector_s {
    ulong_t logical_sect;
    int head, track, sect;
    uchar_t bufr[512];
    saved_sector_ptr_t next;
  };
saved_sector_ptr_t save_sectors;
saved_sector_ptr_t loaded_sectors;


/* save a partition sector into a list of 'em */
void save_sect( saved_sector_ptr_t * sectorlist,
                ulong_t logical_sect, int head, int track, int sect,
                char bufr[512] )
  {
   saved_sector_ptr_t tmp, tmp2;
    /* first, malloc space and fill in the structure: */
    tmp = (saved_sector_ptr_t)malloc( sizeof(saved_sector_t) );
    if (!tmp)
      {
        printf( "ERROR: Out of memory.\n" );
        exit(3);
      }
    tmp->logical_sect = logical_sect;
    tmp->head = head;
    tmp->track = track;
    tmp->sect = sect;
    tmp->next = NULL;
    memcpy( tmp->bufr, bufr, 512 );
    /* then add it to the end of the sector list: */
    if (!*sectorlist)
        *sectorlist = tmp;
    else
      {
        tmp2 = *sectorlist;
        while (tmp2->next)
            tmp2 = tmp2->next;
        tmp2->next = tmp;
      }
  } /* save_sect */


/* print a 512-byte sector dump in hex & ASCII */
void dump_a_sector( char far * bufr )
  {
   int i,j;
    for (i=0; i<512; i++)
      {
        if (0 == (i & 0x0F))
            printf( "  %03X ", i );  /* print offset at start of line */
        printf( " %02X", ((uchar_t*)bufr)[i] );
        if (15 == (i & 0x0F))  /* at end of hex dump line... */
          {
            /* dump line in ASCII, too */
            printf( " " );
            for (j=(i & 0x7F0); j<=i; j++)
              {
                if (0 == (j & 3))
                    printf( " " );
                if ((isgraph(((char*)bufr)[j])) || (' '==((char*)bufr)[j]))
                    printf( "%c", ((char*)bufr)[j] );
                else
                    printf( "" );  /* 0xF9 */
              }
            printf( "\n" );  /* end-of-line */
          }
        else if (3 == (i & 3))
            printf( " " );
      }
  } /* dump_a_sector */


char caption1[] =  "    Beg_CHS_  Sys End_CHS_\n";
char caption2L[] = "__  DH CL CH  Ind DH CL CH  Beg_LBA____  Size_______\n";
/*                  00  01 02 03  04  05 06 07  08 09 10 11  12 13 14 15 */

boolean_t dont_save_more = false_c;


/* write partition data back to hard drive */
void restore_partition_data( int drive, saved_sector_ptr_t sectors )
  {
   int pnum;
   saved_sector_ptr_t olddata;
   saved_sector_ptr_t tmp;
   int rc, rc446, rc_ptable;
   int table_number;

    olddata = save_sectors;
    tmp = sectors;
    table_number = 0;  /* MBR is table_number 1, 1st extended is table_number 2, etc. */
    while (tmp)
      {
        table_number++;
        rc = rc446 = rc_ptable = 0;
        if (olddata)
          {
            rc = memcmp( tmp->bufr, olddata->bufr, 512 );
            rc446 = memcmp( tmp->bufr, olddata->bufr, 446 );  /* part of sector before partition table */
            rc_ptable = memcmp( tmp->bufr+446, olddata->bufr+446, 66 );  /* partition table and 55 AA signature */
          }
        printf( "------------------------------------------------------------------\n" );
        if (olddata && !rc)
            printf( "New partition data is:\n" );
        else
            printf( "New partition data will be:\n" );
        if (rc446)
            dump_a_sector( tmp->bufr );
        fprintf( stdout, "sector=%lu (c=%d, h=%d, s=%d)\n",
                 tmp->logical_sect, tmp->track, tmp->head, tmp->sect );
        partition_logical_sect = tmp->logical_sect;  /* global needed by print_table_entry */
        pt = (struct parttable_s *)(tmp->bufr+0x1BE);
        for (pnum=0; pnum <= 3; pnum++)
            print_table_entry( stdout, table_number, &(pt->part[pnum]) );
        fprintf( stdout, "%02X %02X\n\n",
                 tmp->bufr[510], tmp->bufr[511] );
        printf( "Old partition data is:\n" );
        if (!rc)
          {
            if (!olddata)
                printf( "(non-existant)\n" );
            else
                printf( "Indentical to new data.\n" );
            rc = 0;
          }
        else
          {
            if (!rc_ptable)
                fprintf( stdout, "Not identical, but the partition table portion IS identical.\n" );
            if (rc446)
                dump_a_sector( olddata->bufr );
            if (rc_ptable)
              {
                fprintf( stdout, "sector=%lu (c=%d, h=%d, s=%d)\n",
                         olddata->logical_sect, olddata->track, olddata->head, olddata->sect );
                partition_logical_sect = olddata->logical_sect;  /* global needed by print_table_entry */
                pt = (struct parttable_s *)(olddata->bufr+0x1BE);
                for (pnum=0; pnum <= 3; pnum++)
                    print_table_entry( stdout, table_number, &(pt->part[pnum]) );
                fprintf( stdout, "%02X %02X\n",
                         olddata->bufr[510], olddata->bufr[511] );
                if (tmp->logical_sect != olddata->logical_sect)
                    printf( "ERROR: new and old data are not for same sector!\n" );
              }
          }
        printf( "------------------------------------------------------------------\n" );
        if (rc)
          {
            printf( "Sector to be written on drive %02X is (c=%d, h=%d, s=%d)\n",
                    drive, tmp->track, tmp->head, tmp->sect );
            if (ask("Write new data to hard disk? (data WILL be written immediately if you say Y)"))
              {
                rc = my_disk_wr( drive, tmp->head, tmp->track,
                                 tmp->sect, /*nsects=*/1, tmp->bufr );
                printf( "%s (d=%02X, c=%d, h=%d, s=%d), rc=%d\n",
                         rc ? "Error: Write failed!" : "Sector written",
                         drive, tmp->track, tmp->head, tmp->sect, rc );
              }
          }
        else
          {
            fprintf( stderr, "Press ENTER..." );
            fflush(stderr);  /* force the prompt to display */
            (void)GetChar( /*flush=*/false_c );
            fprintf( stderr, "\n" );
            if (!stderr_equals_stdout())
                printf( "Press ENTER...\n" );
          }
        tmp = tmp->next;
        if (olddata)
            olddata = olddata->next;
      }
  } /* restore_partition_data */


/* save partition data to a file */
void write_partition_data( int drive, saved_sector_ptr_t sectors )
  {
   int bytenum, i, pnum, table_number;
    if (!savefile)
        return;
    if (fixit)
        fprintf( savefile, "----FIXED HARD DISK PARTITION DATA----\n" );
    else
        fprintf( savefile, "----SAVED HARD DISK PARTITION DATA----\n" );
    fprintf( savefile,
            "SHOWPART.EXE v.%s, by Dave Burton <dburton@burtonsys.com>\n\n"
            "Drive = %02X, MaxSector = %d, MaxHead = %d, MaxTrack = %d\n\n",
            version,
            drive, int13h8_maxsector, int13h8_maxhead, int13h8_maxtrack );
    table_number = 0;  /* MBR is table_number 1, 1st extended is table_number 2, etc. */
    while (sectors)
      {
        table_number++;
        fprintf( savefile, "sector=%lu (c=%d, h=%d, s=%d)\n",
                 sectors->logical_sect, sectors->track,
                 sectors->head, sectors->sect );
        /* print bootstrap: */
        for (bytenum=0; bytenum<=0x1BD; bytenum++)
          {
            if (0==(bytenum % 24))
                fprintf( savefile, "%03X  ", bytenum );
            fprintf( savefile, "%02X", sectors->bufr[bytenum] );
            if (23==(bytenum % 24))
              {
                if (bytenum == 0x1AF)
                    fprintf( savefile, "  ;____ending_____" );
                fprintf( savefile, "\n" );
              }
            else
                fprintf( savefile, " " );
          }
        fprintf( savefile, "      ;%s", caption2L );
        /* print partition table proper: */
        partition_logical_sect = sectors->logical_sect;  /* global needed by print_table_entry */
        pt = (struct parttable_s *)(sectors->bufr+0x1BE);
        for (pnum=0; pnum <= 3; pnum++)
          {
            bytenum = 0x1BE+(16*pnum);
            fprintf( savefile, "%03X  ", bytenum );
            for (i=0; i<16; i++)
                fprintf( savefile, "%02X ", sectors->bufr[bytenum+i] );
            fprintf( savefile, ";" );
            print_table_entry( savefile, table_number, &(pt->part[pnum]) );
          }
        fprintf( savefile, "1FE  %02X %02X\n\n",
                 sectors->bufr[510], sectors->bufr[511] );
        sectors = sectors->next;
      }/*while*/
  } /* write_partition_data */


/* Recalculate and fix the start and end head/sect/track values from the
   logical sector numbers, both in the sector location info for each partition
   sector, and also in the four individual partition table slots in each
   partition table sector.  Used for the -F command-line option.  */
boolean_t fix_partition_data( int drive, saved_sector_ptr_t sectors )
  {
   int bytenum, i, pnum, table_number;
   ushort_t track, head, sect, track2, head2, sect2;
   struct partrec_s * prec;
   ulong_t logical_start_sect, logical_end_sect;
   ushort_t sectX, headX, cylX, sectY, headY, cylY;
   boolean_t result;
   boolean_t fixed_this_sector;

    printf( "\ndrive=%02X, sector=1..%d, head=0..%d, track=0..%d\n",
            drive, int13h8_maxsector, int13h8_maxhead, int13h8_maxtrack );
    printf( "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );
    result = false_c;
    table_number = 0;  /* MBR is table_number 1, 1st extended is table_number 2, etc. */
    while (sectors)
      {
        table_number++;
        fixed_this_sector = false_c;
        printf( "sector=%lu (c=%d, h=%d, s=%d)",
                sectors->logical_sect, sectors->track,
                sectors->head, sectors->sect );
        partition_logical_sect = sectors->logical_sect;  /* global needed by print_table_entry */
        calc_T_H_S( sectors->logical_sect, &track, &head, &sect );
        if ((sectors->track != track) || (sectors->head != head) || (sectors->sect != sect))
          {
            result = true_c;
            fixed_this_sector = true_c;
            sectors->track = track;
            sectors->head = head;
            sectors->sect = sect;
            printf( ", changed to (c=%d, h=%d, s=%d)",
                    sectors->logical_sect, track, head, sect );
          }
        printf( "\n" );
        pt = (struct parttable_s *)(sectors->bufr+0x1BE);
        for (pnum=0; pnum <= 3; pnum++)
          {
            prec = &(pt->part[pnum]);
            if (!prec->relative_sect)
                continue;  /* empty (zero) entry in partition table */
            logical_start_sect = calc_logical_start_sect( prec, table_number,
                             partition_logical_sect, start_of_first_extended );
            logical_end_sect = prec->number_of_sect + logical_start_sect - 1LU;
            split_CHSs( prec, &cylX, &headX, &sectX, &cylY, &headY, &sectY );  /* old */
            calc_T_H_S( logical_start_sect, &track, &head, &sect ); /* new */
            calc_T_H_S( logical_end_sect, &track2, &head2, &sect2 ); /* new */
            if (  (track != cylX) || (head != headX) || (sect != sectX)
                || (track2 != cylY) || (head2 != headY) || (sect2 != sectY) )
              {
                result = true_c;
                fixed_this_sector = true_c;
                printf( "Fixing partition table slot #%d, old and new are:\n %s %s ",
                        pnum, caption1, caption2L );
                 print_table_entry( stdout, table_number, prec );
                prec->s.head = head;  /* head | ((track & 0x0C00) >> 4); */
                prec->s.sector = sect | (0xC0 & (track >> 2));
                prec->s.cyl = (track & 0xFF);
                prec->e.head = head2;  /* head2 | ((track2 & 0x0C00) >> 4); */
                prec->e.sector = sect2 | (0xC0 & (track2 >> 2));
                prec->e.cyl = (track2 & 0xFF);
                printf( " " );
                 print_table_entry( stdout, table_number, prec );
                if (sect != 1)
                    printf( " Warning: starting sector #%d is .ne. 1\n", sect );
                if (track > 1023)
                  {
                    printf( " Warning: starting track #%d is .gt. 1023, stored as %d\n",
                            track, (track & 1023) );
                    if (int13h8_maxhead <= 63)
                      if (ask(" Store bits 10-11 of track number into bits 6-7 of head number?\n"))
                        {
                          prec->s.head = head | ((track & 0x0C00) >> 4);
                          if (track > 4095)
                              printf( " Warning: starting track #%d is .gt. 4095, stored as %d\n",
                                      track, (track & 4095) );
                        }
                  }
                if (track2 > 1023)
                  {
                    printf( " Warning: ending track #%d is .gt. 1023, stored as %d\n",
                            track2, (track2 & 1023) );
                    if (int13h8_maxhead <= 63)
                      if (ask(" Store bits 10-11 of track number into bits 6-7 of head number?\n"))
                        {
                          prec->e.head = head2 | ((track2 & 0x0C00) >> 4);
                          if (track2 > 4095)
                              printf( " Warning: ending track #%d is .gt. 4095, stored as %d\n",
                                      track2, (track2 & 4095) );
                        }
                  }
              }
          }/*for*/
        sectors = sectors->next;
        if (!fixed_this_sector)
            printf( "(partition table did not require repair)\n");
        if (sectors)
            printf( "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );
      }/*while*/
    printf( "-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-\n" );
    return result;
  } /* fix_partition_data */


/* return 0 for equal, non-0 for unequal */
int compare_partition_data( saved_sector_ptr_t sectlist1,
                            saved_sector_ptr_t sectlist2 )
  {
   int cmprsize = sizeof(saved_sector_t) - sizeof(saved_sector_ptr_t);
   int rc;
    for (;;)
      {
        if ((!sectlist1) && !sectlist2)
            return 0;
        if (!sectlist2)
            return 1;
        if (!sectlist1)
            return 1;
        rc = memcmp( sectlist1, sectlist2, cmprsize );
        if (rc)
            return 1;
        sectlist1 = sectlist1->next;
        sectlist2 = sectlist2->next;
      }/*for*/
  } /* compare_partition_data */


/* load partition data from already-open loadfile, return true for success */
boolean_t read_partition_data( saved_sector_ptr_t * sectors )
  {
   int bytenum, i, pnum;
   char * rc;
   char * at;
   int rdMaxSector, rdMaxHead, rdMaxTrack, rdTrack, rdHead, rdSector;
   ulong_t rd_logical_sect;
   int bytnum, byteval;
   static uchar_t bufr[512];
   static char line[258];

    if (!loadfile)
        return false_c;

    rc = fgets( line, 256, loadfile );
    if (!strstr(line,"HARD DISK PARTITION DATA----"))
      {
       badfile:
        printf( "ERROR: could not load invalid partition data file\n" );
        return false_c;
      }
    memset( line, 0, 256 );
    while (NULL != (rc=fgets( line, 256, loadfile )))
      {
        if (!strncmp("Drive = ",line,8))
            break;
        memset( line, 0, 256 );
      }
    if (!rc)
        goto badfile;

    at = strstr( line, "MaxSector = " );
    if (!at)
        goto badfile;
    sscanf( at+12, "%d", &rdMaxSector );
    if (int13h8_maxsector != rdMaxSector)
        printf( "Warning!  Loaded file has MaxSector=%d, but current MaxSector=%d\n",
                rdMaxSector, int13h8_maxsector );

    at = strstr( line, "MaxHead = " );
    if (!at)
        goto badfile;
    sscanf( at+10, "%d", &rdMaxHead );
    if (int13h8_maxhead != rdMaxHead)
        printf( "Warning!  Loaded file has MaxHead=%d, but current MaxHead=%d\n",
                rdMaxHead, int13h8_maxhead );

    at = strstr( line, "MaxTrack = " );
    if (!at)
        goto badfile;
    sscanf( at+11, "%d", &rdMaxTrack );
    if (int13h8_maxtrack != rdMaxTrack)
        printf( "Warning!  Loaded file has MaxTrack=%d, but current MaxTrack=%d\n",
                rdMaxTrack, int13h8_maxtrack );

    printf( "dbg: Loaded file has MaxTrack=%d, MaxHead=%d, MaxSector=%d\n",
             rdMaxTrack, rdMaxHead, rdMaxSector );

    for (;;)
      {
        memset( line, 0, 256 );
        while (NULL != (rc=fgets( line, 256, loadfile )))
          {
            if (!strncmp("sector=",line,7))
                break;
            memset( line, 0, 256 );
          }
        if (!rc)
          {
            if (!*sectors)
                goto badfile;
            if (ferror(loadfile) || !feof(loadfile))
                goto badfile;
            return true_c;  /* success */
          }

        /* we found the beginning of a sector description, now we must read it: */

        sscanf( line+7, "%lu", &rd_logical_sect);

        at = strstr( line, "(t=" );
        if (!at)
            goto badfile;
        sscanf( at+3, "%d", &rdTrack );

        at = strstr( line, " h=" );
        if (!at)
            goto badfile;
        sscanf( at+3, "%d", &rdHead );

        at = strstr( line, " s=" );
        if (!at)
            goto badfile;
        sscanf( at+3, "%d", &rdSector );

        memset( bufr, 0, 512 );
        memset( line, 0, 256 );
        while (NULL != (rc=fgets( line, 256, loadfile )))
          {
            if ((!line[0]) || ('\n'==line[0]))
                break;  /* sector ends with a blank line (or end-of-file) */
            if (strlen(line) < 6)
                goto badfile;
            if (('0' != line[0]) && ('1' != line[0]))
                continue;  /* skip lines that don't start with a 3-digit hex byte-offset */
            bytnum = 0;
            sscanf( line, "%x", &bytnum );
            if (debugmode)
              {
                printf("ld:\n%03x ", bytnum );
              }
            at = line+4;
            while (1)
              {
                while (' '==*at)
                    at++;
                if ((';'==*at) || ('\n'==*at) || ('\0'==*at))
                    break;
                if (bytnum >= 512)
                    goto badfile;
                sscanf( at, "%x", &byteval );
                at += 2;
                bufr[bytnum++] = byteval;
                if (debugmode)
                  {
                    printf("%02X ", (0x00FF & byteval) );
                  }
              }
            if (debugmode)
              {
                printf(" ([30h]=%02X)\n", bufr[0x30] );
              }
            memset( line, 0, 256 );
          }
        if (bytnum != 512)
            goto badfile;
        if (debugmode)
          {
            printf("---------- begin dbg dmp ----------\n");
            printf("loaded...  xx,%d,%u,%d):\n", rdHead, rdTrack, rdSector );
            dump_a_sector(bufr);
            printf("----------- end dbg dmp -----------\n");
          }
        save_sect( sectors, rd_logical_sect, rdHead, rdTrack, rdSector, bufr );
      }/*for*/
  } /* read_partition_data */


/* read a specified sector and dump it in Hex and ASCII */
void dump_sector_at_LBA( int drive, ulong_t lba, boolean_t should_be_boot_sect )
  {
   int rc;
   ushort_t head, sect, track;
   char bufr[512];

    calc_T_H_S( lba, &track, &head, &sect );
    rc = my_disk_rd( drive, head, track, sect, /*nsects=*/1, bufr );
    if (rc)
      {
        printf( "Error: Could not read LBA sector %lu (C/H/S=%u/%u/%u), rc=%d\n",
                lba, track, head, sect, rc );
        return;
      }
    if (verbose)
        dump_a_sector( bufr );
    if (should_be_boot_sect && (0xAA55 != *((ushort_t*)(bufr+510))))  /* small endian assumed */
      {
        printf( "Warning: Sector at lba=%lu (C/H/S=%u/%u/%u) doesn't end in 55 AA.\n",
                lba, track, head, sect );
        if (!verbose)
            printf( "You can see the entire dumped sector by using the '-v' [verbose] option.\n" );
      }
    if (verbose)
        printf( "\n" );
  } /* dump_sector_at_LBA */


char question[150];

/* Read and display the drive's partition table which starts at the specified
   head/track/sector.  Return true for success, false for failure.  */
boolean_t display_partition_table( int drive, int table_number,
                                   ushort_t head, ushort_t track,
                                   ushort_t sect, void * bufr )
  {
   int rc, pnum, maxpnum;
   ulong_t sav_partition_logical_sect;
   ushort_t sectX, headX, cylX, sectY, headY, cylY;  /* starting & ending sector/head/cylinder #s */
   static boolean_t dont_ask_again = false_c;
   boolean_t use_LBA;
   ulong_t partitions_to_dump_sector1[4];
   int i;

    if (table_number == 1)
      {
        start_of_first_extended = prev_ptbl_at_LBA = partition_logical_sect = 0L;
        num_boot_indicators_seen = num_partitions = 0;
      }

    partitions_to_dump_sector1[0] = partitions_to_dump_sector1[1] =
     partitions_to_dump_sector1[2] = partitions_to_dump_sector1[3] = 0L;

    this_ptbl_at_cyl = track;  /* store in global */
    this_ptbl_at_head = head;  /*   "   "    "    */
    this_ptbl_at_sect = sect;  /*   "   "    "    */

    prev_ptbl_at_LBA = partition_logical_sect;  /* previous partition table location */
    partition_logical_sect = (sect - 1LU) + (sect_per_cyl() * track)
                             + (int13h8_maxsector * (ulong_t)head);
    sav_partition_logical_sect = partition_logical_sect;

    if (table_number == 2)
        start_of_first_extended = partition_logical_sect;
        /* location of first extended partition table */

    biggest_head = 0;
    biggest_sector = 0;
    biggest_track = 0;
    extended_head = 0;
    extended_sector = 0;
    extended_track = 0;

    rc = my_disk_rd( drive, head, track, sect, /*nsects=*/1, bufr );
    if (rc)
      {
        printf( "ERROR: Could not read partition table from track %d,\n"
                "my_disk_rd(0x%02X,%d,%d,%d,1,), rc=%d\n",
                track, drive, head, track, sect, rc );
        return false_c;
      }
    pt = (struct parttable_s *)(((char *)bufr)+0x1BE);
    partition_table_bad = 0;

    /* Now display the partition table: */
    printf( "Partition table on track %d, at absolute logical sector (LBA) %lu:\n",
            track, partition_logical_sect );
    if (table_number == 1)
      {
        printf( "This is the first sector on the disk, i.e., the Master Boot Record (MBR).\n" );
        printf( "NT's \"disk signature\" is in the four bytes at offset 0x1b8: %02X %02X %02X %02X\n",
                ((uchar_t*)bufr)[0x1b8], ((uchar_t*)bufr)[0x1b9],
                ((uchar_t*)bufr)[0x1bA], ((uchar_t*)bufr)[0x1bB] );
      }


    if (verbose)
      {
        printf( "Sector dump, LBA=%lu:\n", partition_logical_sect );
        dump_a_sector( bufr );
      }

    printf( caption1 );
    printf( caption2L );

    /* find last non-empty partition table entry: */
    for (maxpnum=3; maxpnum >= 1; maxpnum--)
      {
        if (!all_zeros(&(pt->part[maxpnum])))
            break;
      }

    /* step through non-empty partition table entries */
    for (pnum=0; pnum <= maxpnum; pnum++)
      {
        split_CHSs( &(pt->part[pnum]), &cylX, &headX, &sectX, &cylY, &headY, &sectY );
        if (headY > biggest_head)
            biggest_head = headY;
        if (sectY > biggest_sector)
            biggest_sector = sectY;
        if (cylY > biggest_track)
            biggest_track = cylY;
      }

    /* OS/2 hack, check for high DH bits in partition table, and adjust
       geometry to account for OS/2 DOS box bug */
    if (0xAA55 == pt->AA55H)
      {
        if (   (biggest_head > 63)
            && (int13h8_maxhead <= 63)
            && (biggest_head <= (int13h8_maxhead | (int13h8_highDHbits << 6)))
            && !dont_ask_again )
          {
            printf( "Partition table implies modern \"big head number\" geometry,\n"
                    "but Int 13H, Fn 8 didn't show that.  Could be an OS/2 bug.\n" );
            sprintf( question, "Adjust maximum head# from %d to apparent max of %d (work around OS/2 bug)?",
                     int13h8_maxhead,  int13h8_maxhead | (int13h8_highDHbits << 6) );
            if (ask(question))
                int13h8_maxhead |= (int13h8_highDHbits << 6);
            else
                dont_ask_again = true_c;
          }
      }

    /* step through non-empty partition table entries, displaying them */
    sectX = 0;
    for (pnum=0; pnum <= maxpnum; pnum++)
      {
        split_CHSs( &(pt->part[pnum]), &cylX, &headX, &sectX, &cylY, &headY, &sectY );
        if (sectX || cylX || headX
                   || pt->part[pnum].number_of_sect
                   || pt->part[pnum].boot_ind )
          {
            if ((sectX != 1) || (head != 0))
              {
                partition_table_bad = 1;
                printf( "Warning: beginning sector not 1, or beginning head not 0:\n" );
              }
          }
        use_LBA = print_table_entry( stdout, table_number, &(pt->part[pnum]) );
        /* note: print_table_entry also sets table_entry_start_LBA as side-effect result */
        if (is_extd_type( pt->part[pnum].sys_ind ))
          {
            extended_relative_sect = pt->part[pnum].relative_sect;
            if (use_LBA)
              {
                calc_T_H_S( extended_relative_sect+start_of_first_extended,
                            &extended_track, &extended_head, &extended_sector );
             #if 0
                if ((extended_head != headX) || (extended_sector != sectX) || (extended_track != cylX))
                  {
                    printf( "Note: C/H/S from table ignored; using calculated values from LBA.\n"
                            "  Cylinder: %d -> %d,   Head: %d -> %d,   Sector: %d -> %d\n",
                            cylX, extended_track, headX, extended_head, sectX, extended_sector );
                  }
             #endif
              }
            else
              {
                extended_head = headX;
                extended_sector = sectX;
                extended_track =  cylX;
              }
          }
        else if (verbose)
          {
            partitions_to_dump_sector1[pnum] = table_entry_start_LBA;
            if (verbose)
                printf( "    (Verbose mode: We'll dump the sector at lba=%lu)\n",
                         table_entry_start_LBA );
          }
      }

    /* Display partition table signature */
    printf( " %04XH", pt->AA55H );
    if (0xAA55 != pt->AA55H)
        printf( " <---Error!  Should be AA55h!\n" );

    if (maxpnum != 3)
        printf( "    [%d empty partition table slot(s) not shown]", 3-maxpnum );
    printf( "\n" );

    if (pt->part[0].sys_ind == 0x54)  /* DiskManager DDO partition */
      {
        printf( "Warning: This drive is formatted with OnTrack DiskManager.  You must boot\n"
                "with a DiskManager boot diskette to access the drive.\n" );
      }
    else if (pt->part[0].sys_ind == 0x55)  /* EZ Drive partition */
      {
        printf( "Warning: This drive is formatted with StorageSoft/MicroHouse EZ-Drive.\n"
                "You must boot with an EZ-Drive boot diskette to access the drive.\n" );
      }
    if (biggest_track != int13h8_maxtrack)
      {
        if (biggest_track > int13h8_maxtrack)
          {
            if (biggest_track == (1+int13h8_maxtrack))
              {
                if (Didnt_Grow_NumCyl_By1)
                    printf( "Note:" );
                else
                    printf( "Warning:" );
                printf( " biggest track# in partition table equals last track# on drive plus 1.\n" );
              }
            else
              {
                printf( "Error: biggest track# in partition table is larger than last track on drive.\n" );
                partition_table_bad = 1;
              }
            printf( "Biggest track in partition table=%d, but Int13/Fn8 gives last track=%d",
                    biggest_track, unadjusted_int13h8_maxtrack );
            if (unadjusted_int13h8_maxtrack < (DPT_numcyl-1))
                printf( "\nand the DPT gives the last track as %d.\n", (DPT_numcyl-1) );
            else
                printf( ".\n" );
            if (biggest_track == (1+int13h8_maxtrack))
                printf( "This might be due to OnTrack DiskManager's sleight-of-hand.\n" );
          }
        else if (0 == track)
          { /* master boot record (table_number==1) */
            if (Grew_NumCyl_By1 && (biggest_track == (int13h8_maxtrack-1)))
                printf( "Note: The partition table uses only the cylinders that Int13/Fn8 reported\n"
                        "as being available, rather that the 1-larger number that the DPT shows.\n" );
            else
                printf( "Warning: MBR's partition table doesn't use entire drive.\n" );
          }
      }
    if (biggest_sector != int13h8_maxsector)
      {
        printf( "Error: biggest sector# in partition table is unequal to max sector# for drive.\n"
                "Biggest sector# in partition table=%d, But Int 13 gives max sector=%d\n",
                biggest_sector, int13h8_maxsector );
        partition_table_bad = 1;
      }
    if (biggest_head != int13h8_maxhead)
      {
        printf( "Error: biggest head# in partition table is unequal to max head# for drive.\n"
                "Biggest head# in partition table=%d, But Int 13 gives max head=%d\n",
                biggest_head, int13h8_maxhead );
        if (   (biggest_head == 127) && (int13h8_maxhead == 63) && (track == 0)
            && ask("Are you running in a DOS Box under OS/2?"))
          {
            printf( "Okay, this is a known OS/2 quirk, in which Int 13H, AH=8 (Get Drive Parameters)\n"
                    "returns the wrong number of logical heads.\n" );
            int13h8_maxhead = 127;
            printf( "Corrected drive geometry is:\n"
                    "Maxsector = %d, maxhead = %d, maxtrack = %d\n\n",
                    int13h8_maxsector, int13h8_maxhead, int13h8_maxtrack );
            check_for_headnum_truncation( drive );
          }
        else
            partition_table_bad = 1;
      }
    partition_logical_sect = (sect - 1LU) + (sect_per_cyl() * track)
                                    + (int13h8_maxsector * (ulong_t)head);
    if (partition_table_bad)
        printf( ">>>> Partition table appears to be damaged! <<<<\n" );
    if (sav_partition_logical_sect != partition_logical_sect)
      {
        printf( "Error: Use of corrected drive geometry would place THIS partition table\n"
                "at a different logical sector number (%lu instead of %lu).\n",
                partition_logical_sect, sav_partition_logical_sect );
        partition_table_bad = 1;
      }
    if (savefile && (!dont_save_more) && (partition_table_bad))
      {
       char * question;
        if (extended_track)
            question = "Do you wish to include this & subsequent partition table sectors in\n"
                       "the saved partition table file?";
        else
            question = "Do you wish to include this partition table sector in the saved\n"
                       "partition table file?";
        if (!ask(question))
            dont_save_more = true_c;
      }
    if (!dont_save_more)
      {
        save_sect( &save_sectors,
                   partition_logical_sect, head, track, sect, bufr );
      }

    if (extended_track)
      {
       ulong_t computed_absolute_sect, computed_relative_sect,
                     extended_absolute_sect, temp;
        /* recompute sectors/cylinder, in case we adjusted int13h8_maxhead: */
        extended_absolute_sect = extended_relative_sect + start_of_first_extended;
        printf( " *Extended partition is at track %d, head %d, sector %d,\n"
                "  relative logical sector %lu, absolute logical sector %lu\n",
                extended_track, extended_head, extended_sector,
                extended_relative_sect, extended_absolute_sect );
        computed_absolute_sect = (extended_sector - 1LU)
                         + (int13h8_maxsector * (ulong_t)extended_head)
                         + (sect_per_cyl() * extended_track);
        computed_relative_sect = computed_absolute_sect - start_of_first_extended;
        if (computed_relative_sect != extended_relative_sect)
          {
            printf("  Error: computed rel-sect=%lu\n", computed_relative_sect );
            calc_T_H_S( extended_absolute_sect, &extended_track,
                        &extended_head, &extended_sector );
            printf( "  Correcting to adjusted track %d, head %d, sector %d.\n",
                    extended_track, extended_head, extended_sector );
          }
        computed_absolute_sect = (extended_sector - 1LU)
                         + (int13h8_maxsector * (ulong_t)extended_head)
                         + (sect_per_cyl() * extended_track);
        computed_relative_sect = computed_absolute_sect - start_of_first_extended;
        if (computed_relative_sect != extended_relative_sect)
            printf("  ERR: computed rel-sect=%lu\n", computed_relative_sect );
      }
    printf( "\n" );
    if (verbose)
      {
        if (verbose)
            printf( "Dump of sector at LBA %lu (sector immediately following partition table)\n",
                    partition_logical_sect+1 );
        dump_sector_at_LBA( drive, partition_logical_sect+1, /*is_boot=*/false_c );

        for (i=0; i<4; i++)
          {
            if (partitions_to_dump_sector1[i])
              {
                if (verbose)
                    printf( "Dump of sector at LBA %lu (start sector of a logical drive)\n",
                            partitions_to_dump_sector1[i] );
                dump_sector_at_LBA( drive, partitions_to_dump_sector1[i], /*is_boot=*/true_c );
              }
          }
      }
    return true_c;
  } /* display_partition_table */


/* call EBIOS Int 13h, function 48h (get drive parameters).
   (See http://www.microsoft.com/ddk/ddkdocs/win98ddk/storage_60s3.htm) */
int ebios_get_drive_parms(
                  int drive, /* 0x80 for first hard drive, etc.. */
                  ulong_t * cylinders, /* number of cylinders on disk */
                  ushort_t * heads, /* number of heads on disk */
                  ushort_t * sectors, /* number of sectors per track */
                  ulong_t * disk_size /* number of sectors on entire disk drive */ )
  {
   union REGS regs;
   struct SREGS sregs;
   int result;
   struct { /* result buffer */
     ushort_t info_size; /* size of information buffer = 26 */
     ushort_t flags; /* information flags */
     ulong_t cylinders;  /* number of cylinders on disk */
     ulong_t heads; /* number of heads on disk */
     ulong_t sec_per_track; /* number of sectors per track */
     ulong_t sec_total_lsw; /* number of sectors on requested disk (low 32 bits) */
     ulong_t sec_total_msw; /* number of sectors on requested disk (high 32 bits) */
     ushort_t sector_size; /* number of bytes per sector, 512 */
       } bufr;

    if (drive & 0xFF70)
      {
        printf( "ERR: bad parameter to ebios_get_drive_parms(0x%02X,,,,,)\n", drive );
        exit(2);
      }
    if (!extended_BIOS)
      {
        printf( "ERROR: No EBIOS, LBA disk access unavailable\n" );
        exit(2);
      }
    segread( &sregs );
    memset( &regs, 0, sizeof regs );
    regs.h.ah = 0x48;
    regs.h.dl = drive;
    regs.x.si = FP_OFF( &bufr );
    sregs.ds = FP_SEG( &bufr );

    memset( &bufr, 0, sizeof(bufr) );
    bufr.info_size = sizeof(bufr);  /* 26 */

    int86x( 0x13, &regs, &regs, &sregs );
    result = regs.x.cflag ? regs.h.ah : 0;

    if (!result)
      { /* okay, it worked */

        printf( "Int 13h, Fn 48h returned rc=0, flags=%04Xh,\n", bufr.flags );
        printf( "sect/drive=%lX%08lXh", bufr.sec_total_msw, bufr.sec_total_lsw );
        if ((512 == bufr.sector_size) && !bufr.sec_total_msw)
          {
           ulong_t mb, gb, tenths;
            printf( "=%lu", bufr.sec_total_lsw );
            mb = (976U + bufr.sec_total_lsw) / 1953U;  /* aproximate no. of megabytes */
            if (mb < 1000U)
                printf( " (%d MB)", (int)mb );
            else
              {
                mb += 50U;
                gb = mb / 100U;
                tenths = gb % 10U;
                gb = gb / 10U;
                printf(" (%ld.%d GB)", gb, (int)tenths );
              }
          }
        printf( ",\n" );

        if (0 == (bufr.flags & 2))  /* bit 1 */
          { /* if bit 1 is zero then only total number of sectors
               (and bytes/sector) is valid, C/H/S is not valid */
            /* make some reasonable assumptions */
            bufr.heads = 255;
            bufr.sec_per_track = 63;
            bufr.cylinders = 1024;  /* temporary */
            use_LBA_only = true_c;
          }
        else
          {
            printf( "cylinders=%lu", bufr.cylinders );
            printf( " heads=%lu", bufr.heads );
            printf( " sect/tk=%lu ", bufr.sec_per_track );
          }
        printf( "bytes/sect=%d.\n", bufr.sector_size );

        if (0 == (bufr.flags & 1))  /* bit 0 */
          { /* if bit 0 is zero then DMA boundary errors may occur */
            printf( "Warning: DMA boundary errors are possible with this EBIOS.\n" );
          }

        if (   (0xFFFF8000L & bufr.heads) || (0xFFFF8000L & bufr.sec_per_track)
            || bufr.sec_total_msw || (bufr.sector_size != 512) )
          {
            printf( "ERROR: We can't handle this huge drive geometry:\n"
                    "%lu heads, %lu sectors/track, %d bytes/sector, %lX%08lXh total sectors.\n",
                    bufr.heads, bufr.sec_per_track, bufr.sector_size,
                    bufr.sec_total_msw, bufr.sec_total_lsw );
            exit(2);
          }
        if (0 == (bufr.flags & 2))  /* bit 1 */
          { /* if bit 1 is zero then only total number of sectors
               (and bytes/sector) is valid, C/H/S is not valid */
            /* we already set heads and sec_per_track, above */
            bufr.cylinders = bufr.sec_total_lsw / (ulong_t)(bufr.heads * bufr.sec_per_track);
            printf( "Warning: EBIOS Int 13h / Fn 48h couldn't get C/H/S info.\n"
                    "Assuming C/H/S = %lu/254/63\n", bufr.cylinders );
          }
        if (bufr.flags & 4)  /* bit 2 */
          { /* if bit 2 is set, this drive has removable media */
            if (bufr.flags & 16)  /* bit 4 */
                /* if bit 4 is set, then drive has change-line support */
                printf( "Note: this is a removable-media drive.\n" );
            else
                /* if bit 4 is zero, then drive does not have change-line support */
                printf( "Warning: this is a removable-media drive, but with no change-line support.\n" );
            if (bufr.flags & 64)  /* bit 6 */
              { /* if bit 6 is set, then CHS are maximum possible, rather than
                   actual, because there's no media in the drive. */
                printf( "Error: there is no media in the drive!  Aborting...\n" );
                exit(1);
              }
          }
        if (cylinders)
           *cylinders = bufr.cylinders;
        if (heads)
           *heads = (ushort_t)(bufr.heads);
        if (sectors)
           *sectors = (ushort_t)(bufr.sec_per_track);
        if (disk_size)
           *disk_size = bufr.sec_total_lsw;
      }
    return result;
  } /* ebios_get_drive_parms */


void examine_drive( int drive )
  {
   int rc;
   int prev_track;
   int table_number;
   static uchar_t sector_buffer[512];

    printf("Drive #%02XH", drive );
    if ((drive >= 0) && (drive <= 1))
        printf( " (%c:)", 'A'+drive );
    printf( "\n" );
    if (!
          (   ((drive >= 0) && (drive <= 3))
           || ((drive >= 0x80) && (drive <= 0x87)) )
       )
      {
        printf( "ERROR: invalid drive number.\n" );
        exit(1);
      }

    use_LBA_only = false_c;
    int13h48_disksize = 0;

    /* Get drive parameters: */
    rc = my_biosdisk( /*cmd=*/8, drive, 0,0,0,0,NULL );
    if (rc || (0==int13h8_maxsector))
      {
        printf( "ERROR: Could not get drive parameters, Int 13H AH=8, my_biosdisk(8,0x%X,)=%d\n", drive, rc );
        exit(1);
      }

    printf( "Int 13H, AH=8 (Get Drive Parameters) yields:\n" );
    printf( "MaxCylinder = %d, MaxHead = %d, MaxSector = %d, highDHbits = %d\n",
            int13h8_maxtrack, int13h8_maxhead, int13h8_maxsector,
            int13h8_highDHbits );

    extended_BIOS = check_extensions_present_41h( drive );
    printf( "LBA BIOS Extensions %savailable for this drive.\n",
            (extended_BIOS ? "ARE " : "are NOT ") );
    if (extended_BIOS)
      {
       int rc;
        rc = ebios_get_drive_parms( drive,
                          &int13h48_maxtrack, /* number of cylinders on disk, need to decrement by 1 */
                          &int13h48_maxhead, /* number of heads, need to decrement by 1 */
                          &int13h48_maxsector, /* number of sectors per track */
                          &int13h48_disksize );
        if (rc || (!int13h48_maxtrack) || (!int13h48_maxhead) || (!int13h48_maxsector) || (!int13h48_disksize))
          {
            printf( "ERROR: Int 13H, Fn 48h (EBIOS get disk params) returned error code %d\n" );
          }
        else
          {
            printf( "EBIOS Int 13h, Fn 48h (get drive parameters) returned:\n"
                    "%lu cylinders, %u heads, %u sectors/trk (%lu sectors on drive)\n",
                    int13h48_maxtrack, int13h48_maxhead,
                    int13h48_maxsector, int13h48_disksize );
            int13h48_maxtrack--;  /* convert from number of cylinders to max cylinder number */
            int13h48_maxhead--;  /* convert from number of heads to max head number */
            /* sectors within a track are numbered starting with 1, anyhow, so
               there's no conversion */
          }
      }

    {
     ushort_t sv_maxtrack, sv_maxhead, sv_maxsector;
      sv_maxtrack = int13h8_maxtrack;
      sv_maxhead = int13h8_maxhead;
      sv_maxsector = int13h8_maxsector;
      examine_DPT( drive );
      printf( "\n" );
      if (DPT_ptr)
        {
          printf( "Device Parameter Table for drive %02X shows:\n"
                  "NumCylinders = %u, NumHeads = %d, MaxSector = %d\n\n",
                  drive, DPT_numcyl, DPT_numheads, DPT_numsec );
          if (   (sv_maxtrack != int13h8_maxtrack)
              || (sv_maxhead != int13h8_maxhead)
              || (sv_maxsector != int13h8_maxsector) )
            {
              printf( "Revised Drive Parameters are:\n" );
              printf( "MaxCylinder = %d, MaxHead = %d, MaxSector = %d\n\n",
                      int13h8_maxtrack, int13h8_maxhead, int13h8_maxsector );
            }
        }
      else
        {
          printf( "No valid Device Parameter Table found for drive %02X", drive );
          if (!int13h48_disksize)
              printf( "\n" );  /* there's no EBIOS info, either */
          else if (   (int13h48_maxtrack == (ulong_t)int13h8_maxtrack)
                   && ((int)int13h48_maxhead == int13h8_maxhead)
                   && ((int)int13h48_maxsector == int13h8_maxsector) )
              printf( ",\nbut EBIOS drive info is same as Int 13h, Fn 8 info.\n" );
          else
            {
              printf( ",\nbut EBIOS drive info is available.\n" );

              if (   (   ((int)int13h48_maxhead == int13h8_maxhead)
                      || ((int)int13h48_maxhead == (int13h8_maxhead | (int13h8_highDHbits << 6))))
                  && ((int)int13h48_maxsector == int13h8_maxsector)
                  && (int13h48_maxtrack >= (ulong_t)int13h8_maxtrack)
                  && (0L == (int13h48_maxtrack & 0xFFFF8000)) )
                {
                  printf( "MaxCylinder: %d -> %d; Revised Drive Parameters are:\n",
                          int13h8_maxtrack, (int)int13h48_maxtrack );
                  int13h8_maxhead = (int)int13h48_maxhead;
                  int13h8_maxtrack = (int)int13h48_maxtrack;
                  printf( "MaxCylinder = %d, MaxHead = %d, MaxSector = %d\n\n",
                          int13h8_maxtrack, int13h8_maxhead, int13h8_maxsector );
                }
              else
                {
                  printf( "Warning: EBIOS drive parameters inconsistent with old-style geometry:\n"
                          "  Int 13h, Fn 8 (old-style):  C/H/S = %d/%d/%d,\n"
                          "  Int 13h, Fn 48h (EBIOS):    C/H/S = %lu/%u/%u\n"
                          "Using old-style geometry.\n",
                          int13h8_maxtrack, int13h8_maxhead, int13h8_maxsector,
                          int13h48_maxtrack, int13h48_maxhead, int13h48_maxsector );
                }
            }
        }
    }
    if (is_Type3())
        printf( "**** Note: this is an old-style \"BIOS Type 3\" (1025-4096 cyls but <= 65 heads).\n\n" );

    printf("%lu sectors per cylinder\n\n", sect_per_cyl() );

    if (drive < 0x80)
        printf( "Diskettes have no partition table.\n" );
    else
      {
        /* Now read and display MBR (with master partition table): */
        prev_track = 0;
        table_number = 1;
        display_partition_table( drive, table_number, /*head=*/0, /*track=*/0,
                                 /*sect=*/1, sector_buffer );
        /* Now display extended partition(s) */
        while ((extended_track > prev_track) && (extended_track < 17000))
          {
            prev_track = extended_track;
            display_partition_table( drive, ++table_number,
                                     extended_head, extended_track,
                                     extended_sector, sector_buffer );
          }
        if (num_boot_indicators_seen != 1)
          {
            if (num_boot_indicators_seen > 1)
                printf( "Error: %d partitions are \"active\" (BootInd=80H), only one should be.\n",
                        num_boot_indicators_seen );
            else
              {
                /* don't display warning for empty partition table */
                if (num_partitions > 0)
                  {
                    printf( "Warning: no partition is \"active\" (bootable, BootInd=80H).  BIOS might\n"
                            "display \"Operating System not found\" if you attempt to boot from this drive.\n" );
                  }
              }
          }
        if (savefile)
          {
            if (fixit)
                fix_partition_data( drive, save_sectors );
            write_partition_data( drive, save_sectors );
          }
        if (loadfile)
          {
            if (read_partition_data( &loaded_sectors ))
              {
                if (compare_partition_data( save_sectors, loaded_sectors ))
                  {
                    printf( "Loaded partition data is NOT identical to drive.\n" );
                    if (writeit)
                      {
                        printf("\n");
                        if (ask("Write loaded data to hard drive (2nd-to-last chance to chicken out)?"))
                            restore_partition_data( drive, loaded_sectors );
                      }
                  }
                else
                    printf( "Loaded partition data is identical to drive.\n" );
              }
          }
      }
  } /* examine_drive */


int main (int argc, char *argv[])
  {
   int drive, argn;
    printf( "Showpart v.%s - show/rd/wr partition table  ", version );
    printf( "(running DOS %02d.%02d)\n", _osmajor, _osminor );
    printf( copr );
    if (   (argc >= 2)
        && (   ((('-'==argv[1][0])||('/'==argv[1][0])) && ('?'==argv[1][1]))
            || (!stricmp("--help",argv[1]))
           ) )
      {
        argc = 1;  /* "/?" or "-?" or "--help" display help message */
      }
    if (argc < 2)
      {
        printf("\nUsage:\n"
               "   showpart nn [options]\n"
               "   (where nn is the hexadecimal drive number:\n"
               "    00 for A:, 01 for B:, 80 for C:, 81 for 2nd hard drive)\n"
               "\n"
               "Where [options] can be omitted or one of these experimental options:\n"
               "   -V          (verbose, dumps key sectors in hex, too, including boot sectors)\n"
           /*   "   -d          (debug mode -- just for Dave, so don't use this option)\n" -- don't advertise this */
               "   -Sfilename  (where filename is file to which partition info will be Saved)\n"
               "   -Ffilename  (same as -S, but data is Fixed-up for current drive geometry)\n"
               "   -Wfilename  (where filename is partition info file to Write to HDD) !DANGER!\n"
               "   -Tfilename  (where filename is partition info file to Test (compare to HDD)\n"
               "\n"
               "Example (to display C: partition data):\n"
               "   showpart 80 |more\n"
               "Example (to display showpart's help screen):\n"
               "   showpart |more\n"
               "\n"
               "Showpart DISPLAYS hard disk drive partition tables.  If you are daring, you\n"
               "can also use it to EDIT partition tables, by first saving the table to a\n"
               "human-readable text file via the -S option, then editing the saved text file\n"
               "with any text editor, then writing the modified partition table to the hard\n"
               "disk drive with the -W option.  However, I make NO WARRANTY that doing so won't\n"
               "do something horrible.  So use this tool (especially the -W option!) AT YOUR\n"
               "OWN RISK!  Also, note that it DOES NOT fully support drives larger than 137 GB.\n" );

        /* Note: I think it should work on drives that have partitions that
             extend beyond 137.4 GB, but not on drives with partitions that
             begin beyond 137.4 GB, because it doesn't use 48-bit LBA
             addressing.
           see:
             http://www.48bitlba.com/
           and:
             "BIOS Enhanced Disk Drive Services - 3 (EDD-3)":
             http://www.t13.org/docs2004/D1572r2-EDD3.doc
           and:
             "Hard Disk Size Barriers":
             http://www.pcguide.com/ref/hdd/bios/size.htm
           and:
             "The ATA Interface Limit (128 GiB / 137 GB) Barrier":
             http://www.pcguide.com/ref/hdd/bios/size_GB128.htm
           and:
             Microsoft Knowledge Base Article 303013, "How to enable 48-bit
             Logical Block Addressing support for ATAPI disk drives in
             Windows XP":
             http://support.microsoft.com/default.aspx?scid=kb;EN-US;303013
           and:
             Microsoft Knowledge Base Article 305098, "48-Bit LBA Support
             for ATAPI Disk Drives in Windows 2000":
             http://support.microsoft.com/default.aspx?scid=kb;EN-US;305098
           and:
             Intel "48-bit LBA BIOS Support":
             http://support.intel.com/support/chipsets/iaa/sb/CS-009281.htm
        */
               return 1;
      }
    sscanf( argv[1], "%x", &drive );
    for (argn=2; argn<argc; argn++)
      {
        if (!strnicmp(argv[argn],"-T",2))
          {
           dashT:
            if (loadfilename)
                goto badoption;
            loadfilename = argv[argn]+2;
            if (!loadfilename[0])
                goto badoption;
            loadfile = fopen( loadfilename, "rt" );
            if (!loadfile)
              {
                printf( "ERROR: could not read '%s'\n", loadfilename );
                return 1;
              }
          }
        else if (!strncmp(argv[argn],"-W",2))
          {
            writeit = true_c;
            goto dashT;
          }
        else if (!strncmp(argv[argn],"-w",2))
          {
            printf("ERROR: \"-W\" option can't be lower-case.\n");
            goto badoption;
          }
        else if (!strnicmp(argv[argn],"-S",2))
          {
           dashS:
            if (savefilename)
                goto badoption;
            savefilename = argv[argn]+2;
            if (!savefilename[0])
                goto badoption;
            savefile = fopen( savefilename, "wt" );
            if (!savefile)
              {
                printf( "ERROR: could not create '%s'\n", savefilename );
                return 1;
              }
          }
        else if (!strncmp(argv[argn],"-F",2))
          {
            fixit = true_c;
            goto dashS;
          }
        else if (!stricmp(argv[argn],"-V"))
          {
            verbose = true_c;
          }
        else if (!stricmp(argv[argn],"-D"))
          {
            debugmode = true_c;
          }
        else
          {
           badoption:
            printf( "ERROR: invalid option.  Run without parameters for help.\n" );
            return 1;
          }
        if (drive < 0x80)
          {
            printf( "ERROR: Diskettes have no partition table.\n" );
            return 1;
          }
      }/*for*/
    examine_drive( drive );
    if (savefile)
        fclose( savefile );

    return 0;
  } /*main*/
